from __future__ import division
from base import UncertaintyBase
from ..errors import InvalidParamsError
import numpy as np


class StudentsTUncertainty(UncertaintyBase):

    r"""
The Student's T uncertainty distribution probability density function is a function of :math:`\nu`, the degrees of freedom:

.. math:: f(x; \nu) = \frac{\Gamma(\frac{\nu+1}{2})} {\sqrt{\nu\pi}\,\Gamma(\frac{\nu}{2})} \left(1+\frac{x^2}{\nu} \right)^{-\frac{\nu+1}{2}}

A non-standardized distribution, with a location and scale parameter, is also possible, through the transformation:

.. math:: X = \mu + \sigma f

In our implementation, the location parameter :math:`\mu` is ``location``, the scale parameter :math:`\sigma` is ``scale``, and :math:`\nu` (the degrees of freedom)  is ``shape``.

See http://en.wikipedia.org/wiki/Student%27s_t-distribution
    """
    id = 12
    description = "Student's T uncertainty"

    @classmethod
    def validate(cls, params):
        if (params['shape'] <= 0).sum() or np.isnan(params['shape']).sum():
            raise InvalidParamsError(
                "Positive ``nu`` (degrees of freedom) values are required for"
                " Student's T."
            )
        if (params['scale'] <= 0).sum():
            raise InvalidParamsError(
                "Scale values, if specified, must be greater than zero for"
                " Student's T."
            )

    @classmethod
    def random_variables(cls, params, size, seeded_random=None, **kwargs):
        if seeded_random is None:
            seeded_random = np.random
        scale = params['scale'].copy()
        scale[np.isnan(scale)] = 1
        location = params['loc'].copy()
        location[np.isnan(location)] = 0
        data = location + scale * seeded_random.standard_t(
            params['shape'],
            size=(size, params.shape[0])
        ).T
        return data



from .rng import RandomNumberGeneratorTestCase
from .base import UncertaintyTestCase
# from .lhc import LatinHypercubeMCTestCase
from .distributions import *
from .conversions import ConversionTestCase
from .choices import UncertaintyChoicesTestCase



from __future__ import division
from ...distributions import GammaUncertainty
from ...errors import InvalidParamsError
from ..base import UncertaintyTestCase
import numpy as np


class GammaUncertaintyTestCase(UncertaintyTestCase):

    def pretty_close(self, a, b):
        if b == 0:
            self.assertTrue(a - 0.05 < b < a + 0.05)
        else:
            self.assertTrue(0.95 * a < b < 1.05 * a)

    def test_random_variables(self):
        params = self.make_params_array()
        params['shape'] = 2
        params['scale'] = 5
        sample = GammaUncertainty.random_variables(params, 10000)
        # Mean: shape * scale
        self.pretty_close(2 * 5, np.mean(sample))
        # Mean: shape * scale^2
        self.pretty_close(2 * 5 ** 2, np.var(sample))

    def test_random_variables_2d(self):
        params = self.make_params_array(2)
        params['shape'] = (2, 3)
        params['scale'] = (5, 10)
        sample = GammaUncertainty.random_variables(params, 10000)
        self.pretty_close(2 * 5, np.mean(sample[0, :]))
        self.pretty_close(3 * 10, np.mean(sample[1, :]))

    def test_random_variables_offset(self):
        params = self.make_params_array(2)
        params['shape'] = (2, 3)
        params['scale'] = (5, 10)
        params['loc'] = (100, np.NaN)
        sample = GammaUncertainty.random_variables(params, 10000)
        self.pretty_close(2 * 5 + 100, np.mean(sample[0, :]))
        self.pretty_close(3 * 10, np.mean(sample[1, :]))

    def test_loc_nan_ok(self):
        params = self.make_params_array()
        params['loc'] = np.NaN
        params['scale'] = 1
        params['shape'] = 1
        GammaUncertainty.validate(params)
        return True

    def test_scale_validation(self):
        params = self.make_params_array()
        params['loc'] = np.NaN
        params['scale'] = -1
        params['shape'] = 1
        self.assertRaises(
            InvalidParamsError,
            GammaUncertainty.validate,
            params
        )

    def test_shape_validation(self):
        params = self.make_params_array()
        params['loc'] = np.NaN
        params['scale'] = 1
        params['shape'] = -1
        self.assertRaises(
            InvalidParamsError,
            GammaUncertainty.validate,
            params
        )



from ...distributions import UniformUncertainty
from ..base import UncertaintyTestCase
import numpy as np


class UniformTestCase(UncertaintyTestCase):

    def unif_params_1d(self):
        oneDparams = self.make_params_array(1)
        oneDparams['minimum'] = 1
        oneDparams['loc'] = 2
        oneDparams['maximum'] = 3
        return oneDparams

    def unif_params_2d(self):
        params = self.make_params_array(2)
        params['minimum'] = 1
        params['loc'] = 2
        params['maximum'] = 3
        return params

    def test_uniform_ppf(self):
        oneDparams = self.unif_params_1d()
        params = self.unif_params_2d()
        self.assertTrue(np.allclose(np.array([1, 2, 3]),
                                    UniformUncertainty.ppf(oneDparams, np.array([[0, 0.5, 1]]))))
        self.assertTrue(np.allclose(np.array([[1], [2]]),
                                    UniformUncertainty.ppf(params, np.array([0, 0.5]))))

    def test_uniform_cdf(self):
        oneDparams = self.unif_params_1d()
        params = self.unif_params_2d()
        self.assertTrue(np.allclose(np.array([0, 0.5, 1]),
                                    UniformUncertainty.cdf(oneDparams, np.array([[1, 2, 3, ]]))))
        self.assertTrue(np.allclose(np.array([[0], [0.5]]),
                                    UniformUncertainty.cdf(params, np.array([1, 2]))))

    def test_uniform_seeded_random(self):
        oneDparams = self.unif_params_1d()
        self.assertTrue(np.allclose(2.15281272,
                                    UniformUncertainty.random_variables(oneDparams, 1,
                                                                        self.seeded_random())))

    def test_uniform_random(self):
        oneDparams = self.unif_params_1d()
        params = self.unif_params_2d()
        variables = UniformUncertainty.random_variables(oneDparams, size=5000)
        self.assertEqual((1, 5000), variables.shape)
        self.assertTrue(1.95 < np.average(variables) < 2.05)
        variables = UniformUncertainty.random_variables(params, size=5000)
        self.assertEqual((2, 5000), variables.shape)
        self.assertFalse(np.allclose(variables[0, :], variables[1,:]))
        self.assertTrue(1.95 < np.average(variables[0, :]) < 2.05)
        self.assertTrue(1.95 < np.average(variables[1, :]) < 2.05)

    def test_uniform_statistics(self):
        oneDparams = self.unif_params_1d()
        self.assertEqual({'mean': 2, 'mode': 2, 'median': 2, 'lower': 1,
                          'upper': 3}, UniformUncertainty.statistics(oneDparams))

    def test_uniform_pdf(self):
        oneDparams = self.unif_params_1d()
        xs, ys = UniformUncertainty.pdf(oneDparams)
        self.assertTrue(np.allclose(np.array([1, 3]), xs))
        self.assertTrue(np.allclose(np.array([0.5, 0.5]), ys))
        points = np.array([1, 2, 3])
        xs, ys = UniformUncertainty.pdf(oneDparams, points)
        self.assertTrue(np.allclose(points, xs))
        self.assertTrue(np.allclose(np.array([0.5, 0.5, 0.5]), ys))



from .base import UncertaintyBase, BoundedUncertaintyBase
from .undefined import UndefinedUncertainty, NoUncertainty
from .bernoulli import BernoulliUncertainty
from .beta import BetaUncertainty
from .discrete_uniform import DiscreteUniform
from .extreme import GeneralizedExtremeValueUncertainty
from .gamma import GammaUncertainty
from .geometric import UniformUncertainty, TriangularUncertainty
from .lognormal import LognormalUncertainty
from .normal import NormalUncertainty
from .student import StudentsTUncertainty
from .weibull import WeibullUncertainty



# TODO: Write Monte Carlo tests

__version__ = (0, 3, 1)

from .distributions import *
from .uncertainty_choices import uncertainty_choices
from .errors import *
from .random import RandomNumberGenerator, LatinHypercubeRNG, \
    MCRandomNumberGenerator



from ...distributions import TriangularUncertainty
from ..base import UncertaintyTestCase
import numpy as np


class TriangularTestCase(UncertaintyTestCase):

    def test_triangular(self):
        oneDparams = self.biased_params_1d()
        params = self.biased_params_2d()

    def test_triangular_ppf(self):
        oneDparams = self.biased_params_1d()
        params = self.biased_params_2d()
        percentages = np.array([[0, 0.5, 1]])
        self.assertTrue(np.allclose(TriangularUncertainty.ppf(oneDparams,
                                                              percentages), np.array([1, 2.73205083, 4])))
        self.assertTrue(np.allclose(TriangularUncertainty.ppf(params,
                                                              np.array(
                                                                  [[0, 0.5, 1], [0, 0.5, 1]])), np.array(
            [[1, 2.73205083, 4], [1,
                                  2.73205083, 4]])))

    def test_triangular_cdf(self):
        oneDparams = self.biased_params_1d()
        params = self.biased_params_2d()
        percentages = np.array([0, 0.5, 1])
        self.assertTrue(np.allclose(TriangularUncertainty.cdf(oneDparams,
                                                              np.array([[1, 2.73205083, 4]])), percentages))
        self.assertTrue(np.allclose(TriangularUncertainty.cdf(params,
                                                              np.array([1, 2.73205083])), np.array([[0], [0.5]])))

    def test_triangular_seeded_random(self):
        oneDparams = self.biased_params_1d()
        self.assertTrue(np.allclose(2.85968765,
                                    TriangularUncertainty.random_variables(oneDparams, 1,
                                                                           self.seeded_random())))

    def test_triangular_random(self):
        oneDparams = self.biased_params_1d()
        params = self.biased_params_2d()
        variables = TriangularUncertainty.random_variables(oneDparams, 5000)
        self.assertTrue(2.61 < np.average(variables) < 2.71)
        self.assertEqual((1, 5000), variables.shape)
        variables = TriangularUncertainty.random_variables(params, 5000)
        self.assertFalse(np.allclose(variables[0, :], variables[1,:]))
        self.assertFalse(np.allclose(variables[0, :], variables[1,:]))
        self.assertTrue(2.61 < np.average(variables[0, :]) < 2.71)
        self.assertTrue(2.61 < np.average(variables[1, :]) < 2.71)

    def test_triangular_statistics(self):
        oneDparams = self.biased_params_1d()
        tri_stats = {'upper': None, 'lower': None, 'median': 2.732050895690918,
                     'mode': 3.0, 'mean': 2.6666667461395264}
        self.assertEqual(TriangularUncertainty.statistics(oneDparams),
                         tri_stats)

    def test_triangular_pdf(self):
        oneDparams = self.biased_params_1d()
        xs, ys = TriangularUncertainty.pdf(oneDparams)
        self.assertTrue(np.allclose(np.array([1, 3, 4]), xs))
        self.assertTrue(np.allclose(np.array([0, 0.66666669, 0]), ys))
        points = np.array([1, 2, 3, 4])
        xs, ys = TriangularUncertainty.pdf(oneDparams, points)
        self.assertTrue(np.allclose(points, xs))
        self.assertTrue(np.allclose(np.array([0, 0.99999997, 1.99999994, 0]), ys))



import unittest
from numpy import *
from ..random import RandomNumberGenerator as RNG
from ..distributions import *
from ..errors import UnknownUncertaintyType
import numpy as np


class RandomNumberGeneratorTestCase(unittest.TestCase):

    def test_invalid_uncertainty_type(self):
        with self.assertRaises(UnknownUncertaintyType):
            RNG(object, UncertaintyBase.from_dicts({}))

    def test_uncertainty_not_subclass(self):
        class Foo(object):

            @classmethod
            def bounded_random_variables(self):
                pass

            @classmethod
            def validate(self, *args, **kwargs):
                pass

        RNG(Foo, UncertaintyBase.from_dicts({}))

    def test_method_call(self):
        rng = RNG(
            NormalUncertainty,
            UncertaintyBase.from_dicts({'loc': 0, 'scale': 1})
        )
        rng.next()
        rng.generate_random_numbers()

    def test_as_iterator(self):
        counter = 0
        rng = RNG(
            NormalUncertainty,
            UncertaintyBase.from_dicts({'loc': 0, 'scale': 1})
        )
        for x in rng:
            counter += 1
            if counter >= 10:
                break

    def test_seed(self):
        data = []
        for x in range(2):
            rng = RNG(
                NormalUncertainty,
                UncertaintyBase.from_dicts({'loc': 0, 'scale': 1}),
                seed=111
            )
            data.append(rng.next())
        self.assertTrue(np.allclose(*data))

    def test_output_dimensions(self):
        rng = RNG(
            NormalUncertainty,
            UncertaintyBase.from_dicts(
                {'loc': 0, 'scale': 1},
                {'loc': 1, 'scale': 2}
            ),
            size=10
        )
        self.assertTrue(rng.next().shape, (2, 10))



import unittest
from ..distributions import *
from ..errors import ImproperBoundsError, \
    UndefinedDistributionError, InvalidParamsError, UnreasonableBoundsError
from nose.plugins.skip import SkipTest
import numpy as np


class UncertaintyTestCase(unittest.TestCase):

    def make_params_array(self, length=1):
        assert isinstance(length, int)
        params = np.zeros((length,), dtype=[
            ('input', 'u4'),
            ('output', 'u4'),
            ('loc', 'f4'),
            ('negative', 'b1'),
            ('scale', 'f4'),
            ('shape', 'f4'),
            ('minimum', 'f4'),
            ('maximum', 'f4')
        ])
        params['minimum'] = params['maximum'] = np.NaN
        params['scale'] = params['shape'] = np.NaN
        return params

    def seeded_random(self, seed=111111):
        return np.random.RandomState(seed)

    def biased_params_1d(self):
        oneDparams = self.make_params_array(1)
        oneDparams['minimum'] = 1
        oneDparams['loc'] = 3
        oneDparams['maximum'] = 4
        return oneDparams

    def biased_params_2d(self):
        params = self.make_params_array(2)
        params['minimum'] = 1
        params['loc'] = 3
        params['maximum'] = 4
        return params


class BaseTestCase(UncertaintyTestCase):

    def test_uncertainty_base_validate(self):
        """UncertaintyBase: Mean exists, and bounds are ok if present."""
        params = self.make_params_array(1)
        params['maximum'] = 2
        params['minimum'] = 2.1
        self.assertRaises(
            ImproperBoundsError,
            UncertaintyBase.validate,
            params
        )

    def test_check_2d_inputs(self):
        params = self.make_params_array(2)
        params['minimum'] = 0
        params['loc'] = 1
        params['maximum'] = 2
        # Params has 2 rows. The input vector can only have shape (2,) or (2, n)
        self.assertRaises(
            InvalidParamsError,
            UncertaintyBase.check_2d_inputs,
            params,
            np.array((1,))
        )
        self.assertRaises(
            InvalidParamsError,
            UncertaintyBase.check_2d_inputs,
            params,
            np.array(((1, 2),))
        )
        self.assertRaises(
            InvalidParamsError,
            UncertaintyBase.check_2d_inputs,
            params,
            np.array(((1, 2), (3, 4), (5, 6)))
        )
        # Test 1-d input
        vector = UncertaintyBase.check_2d_inputs(params, np.array((1, 2)))
        self.assertTrue(np.allclose(vector, np.array(([1], [2]))))
        # Test 1-row 2-d input
        vector = UncertaintyBase.check_2d_inputs(
            params,
            np.array(((1, 2, 3), (1, 2, 3)))
        )
        self.assertTrue(np.allclose(vector, np.array(((1, 2, 3), (1, 2, 3)))))

    @SkipTest
    def test_check_bounds_reasonableness(self):
        params = self.make_params_array(1)
        params['maximum'] = -0.3
        params['loc'] = 1
        params['scale'] = 1
        self.assertRaises(
            UnreasonableBoundsError,
            NormalUncertainty.check_bounds_reasonableness,
            params
        )

    def test_bounded_random_variables(self):
        params = self.make_params_array(1)
        params['maximum'] = -0.2  # Only ~ 10 percent of distribution
        params['loc'] = 1
        params['scale'] = 1
        sample = NormalUncertainty.bounded_random_variables(
            params,
            size=50000,
            maximum_iterations=1000
        )
        self.assertEqual((sample > -0.2).sum(), 0)
        self.assertEqual(sample.shape, (1, 50000))
        self.assertTrue(np.abs(sample.sum()) > 0)

    def test_bounded_uncertainty_base_validate(self):
        """BoundedUncertaintyBase: Make sure legitimate bounds are provided"""
        params = self.make_params_array(1)
        # Only maximum
        params['maximum'] = 1
        params['minimum'] = np.NaN
        self.assertRaises(
            ImproperBoundsError,
            BoundedUncertaintyBase.validate,
            params
        )
        # Only minimum
        params['maximum'] = np.NaN
        params['minimum'] = -1
        self.assertRaises(
            ImproperBoundsError,
            BoundedUncertaintyBase.validate,
            params
        )

    def test_undefined_uncertainty(self):
        params = self.make_params_array(1)
        self.assertRaises(
            UndefinedDistributionError,
            UndefinedUncertainty.cdf,
            params,
            np.random.random(10)
        )
        params = self.make_params_array(2)
        params['loc'] = 9
        self.assertTrue(np.allclose(
            np.ones((2, 3)) * 9,
            UndefinedUncertainty.random_variables(params, 3)
        ))
        random_percentages = np.random.random(20).reshape(2, 10)
        self.assertTrue(np.allclose(
            np.ones((2, 10)) * 9,
            UndefinedUncertainty.ppf(params, random_percentages)
        ))



from __future__ import division
from base import UncertaintyBase
from ..errors import InvalidParamsError
import numpy as np


class WeibullUncertainty(UncertaintyBase):
    r"""
The Weibull distribution has the probability distribution function:

.. math:: f(x; k, \lambda) = \frac{k}{\lambda} \left( \frac{x}{\lambda} \right)^{k - 1} e^{- \left( x / \lambda \right)^{k}}

In our implementation, :math:`\lambda` is ``scale``, and :math:`k`  is ``shape``. An optional location parameter, which offsets the distribution from the origin, can be specified in ``loc``.

See https://en.wikipedia.org/wiki/Weibull_distribution.
    """
    id = 8
    description = "Weibull uncertainty"

    @classmethod
    def validate(cls, params):
        if (params['shape'] <= 0).sum():
            raise InvalidParamsError(
                "Real, positive ``k`` values need for Weibull.")
        if (params['scale'] <= 0).sum():
            raise InvalidParamsError(
                "Real, positive ``lambda`` values need for Weibull.")

    @classmethod
    def random_variables(cls, params, size, seeded_random=None,
                         transform=False):
        if seeded_random is None:
            seeded_random = np.random
        offset = params['loc'].copy()
        offset[np.isnan(offset)] = 0
        data = offset.reshape((-1, 1)) + \
            params['scale'].reshape((-1, 1)) * \
            seeded_random.weibull(
                params['shape'],
                size=(size, params.shape[0])
            ).T
        data[params['negative'], :] = -1 * data[params['negative'], :]
        return data



from ...distributions import NormalUncertainty
from ..base import UncertaintyTestCase
import numpy as np
from ...errors import InvalidParamsError


class NormalTestCase(UncertaintyTestCase):

    def test_normal_validate(self):
        params = self.make_params_array(1)
        params['scale'] = np.NaN
        self.assertRaises(
            InvalidParamsError,
            NormalUncertainty.validate,
            params
        )

    def test_normal_ppf(self):
        oneDparams = self.biased_params_1d()
        params = self.biased_params_2d()
        oneDparams['scale'] = 0.8
        params['scale'] = 0.8
        self.assertTrue(np.allclose(
            np.array([[1.97475873, 3, 3.673297]]),
            NormalUncertainty.ppf(
                oneDparams,
                np.array([[0.1, 0.5, 0.8]])
            )
        ))
        self.assertTrue(np.allclose(
            np.array([[1.97475873], [3]]),
            NormalUncertainty.ppf(
                params,
                np.array([0.1, 0.5])
            )
        ))

    def test_normal_cdf(self):
        oneDparams = self.biased_params_1d()
        params = self.biased_params_2d()
        oneDparams['scale'] = 0.8
        params['scale'] = 0.8
        self.assertTrue(np.allclose(np.array([[0.00620967, 0.10564978, 0.5,
                                               0.89435022]]),
                                    NormalUncertainty.cdf(oneDparams,
                                                          np.array([[1, 2, 3, 4]]))))
        self.assertTrue(np.allclose(np.array([[0.00620967, 0.10564978, 0.5,
                                               0.89435022]]),
                                    NormalUncertainty.cdf(params, np.array([[1, 2, 3, 4],
                                                                           [1, 2, 3, 4]]))[1, :]))

    def test_normal_seeded_random(self):
        oneDparams = self.biased_params_1d()
        oneDparams['scale'] = 0.8
        self.assertTrue(np.allclose(np.array([[1.49734064]]),
                                    NormalUncertainty.random_variables(oneDparams, 1,
                                                                       self.seeded_random())))

    def test_normal_random(self):
        oneDparams = self.biased_params_1d()
        params = self.biased_params_2d()
        oneDparams['scale'] = 0.8
        params['scale'] = 0.8
        variables = NormalUncertainty.random_variables(oneDparams, size=50000)
        self.assertEqual((1, 50000), variables.shape)
        self.assertTrue(2.95 < np.average(variables) < 3.05)
        self.assertTrue(2.95 < np.median(variables) < 3.05)
        variables = NormalUncertainty.random_variables(params, size=50000)
        self.assertEqual((2, 50000), variables.shape)
        self.assertFalse(np.allclose(variables[0, :], variables[1,:]))
        self.assertTrue(2.95 < np.average(variables[1, :]) < 3.05)
        self.assertTrue(2.95 < np.median(variables[1, :]) < 3.05)
        self.assertTrue(2.95 < np.average(variables[0, :]) < 3.05)

    def test_normal_statistics(self):
        oneDparams = self.biased_params_1d()
        oneDparams['scale'] = 0.8
        self.assertEqual({'upper': 4.5999999046325684,
                          'lower': 1.3999999761581421, 'median': 3.0, 'mode': 3.0,
                          'mean': 3.0},
                         NormalUncertainty.statistics(oneDparams))

    def test_normal_bounded_pdf(self):
        oneDparams = self.biased_params_1d()
        oneDparams['scale'] = 0.8
        points = np.array([[1, 2, 3, 4]])
        xs, ys = NormalUncertainty.pdf(oneDparams.copy(), points)
        self.assertTrue(np.allclose(points, xs))
        self.assertTrue(np.allclose(np.array([0.02191038, 0.22831136, 0.49867784,
                                              0.22831136]), ys))
        xs, ys = NormalUncertainty.pdf(oneDparams.copy())
        self.assertEqual(xs.shape, (200,))
        self.assertEqual(ys.shape, (200,))
        self.assertEqual(xs.min(), 1)
        self.assertTrue(3.98 < xs.max() <= 4)
        self.assertEqual(ys.min(), 0.021910377331033407)
        self.assertTrue(np.allclose(ys.max(), 0.498668095951))

    def test_normal_unbounded_pdf(self):
        oneDparams = self.biased_params_1d()
        oneDparams['scale'] = 0.8
        oneDparams['minimum'] = oneDparams['maximum'] = np.NaN
        xs, ys = NormalUncertainty.pdf(oneDparams.copy())
        self.assertEqual(xs.shape, (200,))
        self.assertEqual(ys.shape, (200,))
        self.assertTrue(np.allclose(xs.min(), 1.23999989033))
        self.assertTrue(np.allclose(xs.max(), 4.74241173267))
        self.assertTrue(np.allclose(np.average(xs), 2.9912058115))
        self.assertTrue(np.allclose(ys.min(), 0.0443432302212))
        self.assertTrue(np.allclose(ys.max(), 0.498677843058))
        self.assertTrue(np.allclose(np.average(ys), 0.276188653511))



from ...distributions import BernoulliUncertainty
from ..base import UncertaintyTestCase
import numpy as np


class BernoulliTestCase(UncertaintyTestCase):

    def test_bernoulli_ppf(self):
        oneDparams = self.biased_params_1d()
        params = self.biased_params_2d()
        self.assertTrue(np.allclose(
            BernoulliUncertainty.ppf(
                oneDparams,
                np.array([[0, 0.25, 0.5, 0.75, 1]])
            ),
            np.array([[1, 1, 1, 4, 4]])
        ))
        self.assertTrue(np.allclose(
            BernoulliUncertainty.ppf(
                params,
                np.array([0.5, 0.8])
            ),
            np.array([[1], [4]])
        ))

    def test_bernoulli_cdf(self):
        oneDparams = self.biased_params_1d()
        params = self.biased_params_2d()
        self.assertTrue(np.allclose(
            BernoulliUncertainty.cdf(
                oneDparams,
                np.array([[1, 2, 3, 4]])
            ),
            np.array([[0, 0, 1, 1]])
        ))
        self.assertTrue(np.allclose(
            BernoulliUncertainty.cdf(
                params,
                np.array([1, 3])
            ),
            np.array([[0], [1]])
        ))

    def test_bernoulli_seeded_random(self):
        oneDparams = self.biased_params_1d()
        print BernoulliUncertainty.random_variables(
            oneDparams,
            10,
            self.seeded_random()
        )
        self.assertTrue(np.allclose(
            BernoulliUncertainty.random_variables(
                oneDparams,
                10,
                self.seeded_random()
            ),
            BernoulliUncertainty.random_variables(
                oneDparams,
                10,
                self.seeded_random()
            )
        ))

    def test_bernoulli_random(self):
        oneDparams = self.biased_params_1d()
        params = self.biased_params_2d()
        variables = BernoulliUncertainty.random_variables(oneDparams, 50000)
        self.assertTrue(0.3 < np.average(variables) < 0.35)
        self.assertEqual(variables.shape, (1, 50000))
        variables = BernoulliUncertainty.random_variables(params, 50000)
        self.assertTrue(0.3 < np.average(variables[0, :]) < 0.35)
        self.assertTrue(0.3 < np.average(variables[1, :]) < 0.35)
        self.assertEqual(variables.shape, (2, 50000))
        self.assertFalse(np.allclose(variables[0, :], variables[1,:]))

    def test_bernoulli_statistics(self):
        oneDparams = self.biased_params_1d()
        bern_stats = {'upper': None, 'lower': None, 'median': None,
                      'mode': None, 'mean': 3}
        self.assertEqual(BernoulliUncertainty.statistics(oneDparams),
                         bern_stats)

    def test_bernoulli_pdf(self):
        oneDparams = self.biased_params_1d()
        self.assertRaises(NotImplementedError, BernoulliUncertainty.pdf,
                          oneDparams)



import unittest
from ..uncertainty_choices import *


class UncertaintyChoicesTestCase(unittest.TestCase):

    def test_contains(self):
        self.assertTrue(UndefinedUncertainty in uncertainty_choices)



from __future__ import division
from ..utils import one_row_params_array
from .base import BoundedUncertaintyBase
from scipy import stats
import numpy as np


class UniformUncertainty(BoundedUncertaintyBase):

    """Continuous uniform distribution. In SciPy, the uniform distribution is defined from loc to loc+scale."""
    id = 4
    description = "Uniform uncertainty"

    @classmethod
    def random_variables(cls, params, size, seeded_random=None):
        if not seeded_random:
            seeded_random = np.random
        return seeded_random.uniform(
            params['minimum'],  # Minimum (low)
            params['maximum'],  # Maximum (high)
            size=(size, params.shape[0])).T

    @classmethod
    def cdf(cls, params, vector):
        vector = cls.check_2d_inputs(params, vector)
        results = np.zeros(vector.shape)
        for row in range(params.shape[0]):
            results[row, :] = stats.uniform.cdf(vector[row, :],
                                                loc=params[row]['minimum'], scale=params[row]['maximum'] -
                                                params[row]['minimum'])
        return results

    @classmethod
    def ppf(cls, params, percentages):
        percentages = cls.check_2d_inputs(params, percentages)
        scale = (params['maximum'] - params['minimum']).reshape(
            params.shape[0], 1)
        return percentages * scale + params['minimum'].reshape(
            params.shape[0], 1)

    @classmethod
    @one_row_params_array
    def statistics(cls, params):
        mean = (params['maximum'] + params['minimum']) / 2
        return {'mean': mean, 'mode': mean, 'median': mean,
                'lower': params['minimum'], 'upper': params['maximum']}

    @classmethod
    @one_row_params_array
    def pdf(cls, params, xs=None):
        if xs == None:
            xs = (params['minimum'], params['maximum'])
        percentage = 1 / (params['maximum'] - params['minimum'])
        ys = np.array([float(percentage) for x in xs])
        return np.array([float(x) for x in xs]), ys


class TriangularUncertainty(BoundedUncertaintyBase):
    id = 5
    description = "Triangular uncertainty"

    @classmethod
    def random_variables(cls, params, size, seeded_random=None):
        if not seeded_random:
            seeded_random = np.random
        return seeded_random.triangular(
            params['minimum'],  # Left
            params['loc'],  # Mode
            params['maximum'],  # Right
            size=(size, params.shape[0])).T

    @classmethod
    def cdf(cls, params, vector):
        vector = cls.check_2d_inputs(params, vector)
        # Adjust parameters to (0,1) range
        adjusted_means, scale = cls.rescale(params)
        # To be broadcasted correctly, scale and mins must be column vectors
        scale.resize(scale.shape[0], 1)
        mins = np.array(params['minimum'])
        mins.resize(params.shape[0], 1)
        # Adjust values to use cdf for to (0,1) range
        adjusted_vector = (vector - mins) / scale
        results = np.zeros(vector.shape)
        for row in range(params.shape[0]):
            results[row, :] = stats.triang.cdf(adjusted_vector[row, :],
                                               adjusted_means[row])
        return results

    @classmethod
    def ppf(cls, params, percentages):
        percentages = cls.check_2d_inputs(params, percentages)
        adjusted_means, scale = cls.rescale(params)
        scale.resize(scale.shape[0], 1)
        adjusted_means.resize(scale.shape[0], 1)
        mins = np.array(params['minimum'])
        mins.resize(params.shape[0], 1)
        return stats.triang.ppf(percentages, adjusted_means) * scale + mins

    @classmethod
    @one_row_params_array
    def statistics(cls, params):
        mode = params['loc']
        mean = (params['minimum'] + params['maximum'] + mode) / 3
        upper = params['maximum']
        lower = params['minimum']
        if mode > (upper - lower) / 2:
            median = lower + ((upper - lower) * (mode - lower)) ** 0.5 / (
                2 ** 0.5)
        elif mode < (upper - lower) / 2:
            median = upper - ((upper - lower) * (upper - mode)
                              ) ** 0.5 / (2 ** 0.5)
        else:
            median = mode
        return {'mean': float(mean), 'median': float(median),
                'mode': float(mode), 'lower': None, 'upper': None}

    @classmethod
    @one_row_params_array
    def pdf(cls, params, xs=None):
        if xs == None:
            lower = params['minimum']
            upper = params['maximum']
            mode = params['loc']
            if not mode:
                mode = (upper + lower) / 2
            xs = np.array([float(x) for x in (lower, mode, upper)])
            ys = np.array([0, float((mode - lower) / (upper - lower)), 0])
        else:
            adjusted_means, scale = cls.rescale(params)
            adj_xs = (xs - params['minimum']) / scale
            ys = stats.triang.pdf(adj_xs, adjusted_means)
        return xs, ys



from .bernoulli import BernoulliTestCase
from .beta import BetaTestCase
from .discrete_uniform import DiscreteUniformTestCase
from .extreme import GeneralizedExtremeValueUncertaintyTestCase
from .gamma import GammaUncertaintyTestCase
from .lognormal import LognormalTestCase
from .normal import NormalTestCase
from .student import StudentsTTestCase
from .triangular import TriangularTestCase
from .uniform import UniformTestCase
from .weibull import WeibullTestCase



from __future__ import division
from ..errors import InvalidParamsError
from ..utils import one_row_params_array
from base import UncertaintyBase
from numpy import random, zeros, isnan, arange
from scipy import stats


class BetaUncertainty(UncertaintyBase):

    r"""
The Beta distribution has the probability distribution function:

.. math:: f(x; \alpha, \beta) = \frac{1}{B(\alpha, \beta)} x^{\alpha - 1}(1 - x)^{\beta - 1},

where the normalisation, *B*, is the beta function:

.. math:: B(\alpha, \beta) = \int_0^1 t^{\alpha - 1}(1 - t)^{\beta - 1} dt

The :math:`\alpha` parameter is ``loc``, and :math:`\beta` is ``shape``. By default, the Beta distribution is defined from 0 to 1; the upper bound can be rescaled with the ``scale`` parameter.

Wikipedia: `Beta distribution <http://en.wikipedia.org/wiki/Beta_distribution>`_
    """
    id = 10
    description = "Beta uncertainty"

    @classmethod
    def validate(cls, params):
        if (params['loc'] > 0).sum() != params.shape[0]:
            raise InvalidParamsError("Real, positive alpha values are" +
                                     " required for Beta uncertainties.")
        if (params['shape'] > 0).sum() != params.shape[0]:
            raise InvalidParamsError("Real, positive beta values are" +
                                     " required for Beta uncertainties.")
        if (params['scale'] <= 0).sum():
            raise InvalidParamsError("Scale value must be positive or NaN")

    @classmethod
    def random_variables(cls, params, size, seeded_random=None,
                         transform=False):
        if not seeded_random:
            seeded_random = random
        scale = params['scale']
        scale[isnan(scale)] = 1
        return scale.reshape((-1, 1)) * seeded_random.beta(
            params['loc'],
            params['shape'],
            size=(size, params.shape[0])).T

    @classmethod
    def cdf(cls, params, vector):
        vector = cls.check_2d_inputs(params, vector)
        results = zeros(vector.shape)
        scale = params['scale']
        scale[isnan(scale)] = 1
        for row in range(params.shape[0]):
            results[row, :] = stats.beta.cdf(vector[row, :],
                                             params['loc'][row], params['shape'][row],
                                             scale=scale[row])
        return results

    @classmethod
    def ppf(cls, params, percentages):
        percentages = cls.check_2d_inputs(params, percentages)
        results = zeros(percentages.shape)
        scale = params['scale']
        scale[isnan(scale)] = 1
        for row in range(percentages.shape[0]):
            results[row, :] = stats.beta.ppf(percentages[row, :],
                                             params['loc'][row], params['shape'][row],
                                             scale=scale[row])
        return results

    @classmethod
    @one_row_params_array
    def statistics(cls, params):
        alpha = float(params['loc'])
        beta = float(params['shape'])
        # scale = 1 if isnan(params['maximum'])[0] else float(params['maximum'])
        if alpha <= 1 or beta <= 1:
            mode = "Undefined"
        else:
            mode = (alpha - 1) / (alpha + beta - 2)
        return {
            'mean': alpha / (alpha + beta),
            'mode': mode,
            'median': "Not Implemented",
            'lower': "Not Implemented",
            'upper': "Not Implemented"
        }

    @classmethod
    @one_row_params_array
    def pdf(cls, params, xs=None):
        scale = 1 if isnan(params['scale'])[0] else float(params['scale'])
        if xs == None:
            xs = arange(0, scale, scale / cls.default_number_points_in_pdf)
        ys = stats.beta.pdf(xs, params['loc'], params['shape'],
                            scale=scale)
        return xs, ys.reshape(ys.shape[1])



from __future__ import division
from ...distributions import DiscreteUniform
from ..base import UncertaintyTestCase
import numpy as np


class DiscreteUniformTestCase(UncertaintyTestCase):

    def test_array_shape_1d(self):
        params = self.make_params_array(length=1)
        params['minimum'] = 0
        params['maximum'] = 10
        sample = DiscreteUniform.random_variables(params, 100)
        self.assertEqual(sample.shape, (1, 100))

    def test_array_shape_2d(self):
        params = self.make_params_array(length=10)
        params['minimum'] = 0
        params['maximum'] = 10
        sample = DiscreteUniform.random_variables(params, 100)
        self.assertEqual(sample.shape, (10, 100))

    def test_random_variables(self):
        params = self.make_params_array(length=10)
        params['minimum'] = 5
        params['maximum'] = 10
        sample = DiscreteUniform.random_variables(params, 10000)
        self.assertEqual([5, 6, 7, 8, 9], np.unique(sample).tolist())



# Weighted formulas from
# http://en.wikipedia.org/wiki/Weighted_mean#Weighted_sample_variance

# TODO: Test against http://pygsl.sourceforge.net/README.html
# http://pygsl.sourceforge.net/reference/pygsl/node36.html


def weighted_mean(values, weights):
    assert values.shape == weights.shape
    return (values * weights).sum() / weights.sum()


def weighted_sample_variance(values, weights):
    assert values.shape == weights.shape
    return ((weights * values ** 2).sum() * weights.sum() -
            ((weights * values).sum()) ** 2) / ((weights.sum()) ** 2 -
                                                (weights ** 2).sum())


def weighted_sample_stddev(values, weights):
    return weighted_sample_variance(values, weights) ** 0.5


def test():
    from numpy import array
    a = array((1, 2, 3, 4))
    print a ** 2
    b = array((0.5, 0.5, 1, 1))
    print weighted_mean(a, b)
    print weighted_sample_stddev(a, b)

if __name__ == '__main__':
    test()



import unittest
from ..distributions import UncertaintyBase
from ..utils import construct_params_array
import numpy as np


class ConversionTestCase(unittest.TestCase):

    def get_right_1d_array(self):
        params = construct_params_array(include_type=True)
        params['loc'] = 1
        params['maximum'] = 2
        params['minimum'] = 0
        params['scale'] = 3
        return params

    def get_right_2d_array(self):
        params = construct_params_array(2, include_type=True)
        params['loc'] = (1, 2)
        params['scale'] = (3, 4)
        params['maximum'] = 10
        params['minimum'] = 0
        return params

    def test_1d_dict(self):
        values = {"loc": 1.0,
                  "scale": 3.0,
                  "maximum": 2.0,
                  "minimum": 0.0,
                  "negative": False,
                  }
        self.sa_allclose(
            UncertaintyBase.from_dicts(values),
            self.get_right_1d_array(),
        )

    def test_1d_tuple(self):
        values = (1, 3, np.NaN, 0, 2, False, 0)
        self.sa_allclose(
            UncertaintyBase.from_tuples(values),
            self.get_right_1d_array(),
        )

    def test_2d_dict(self):
        values = (
            {"loc": 1.0,
             "scale": 3.0,
             "maximum": 10.0,
             "minimum": 0.0,
             "negative": False},
            {"loc": 2.0,
             "scale": 4.0,
             "maximum": 10.0,
             "minimum": 0.0,
             "negative": False}
        )
        self.sa_allclose(
            UncertaintyBase.from_dicts(*values),
            self.get_right_2d_array(),
        )

    def test_2d_tuple(self):
        values = (
            (1, 3, np.NaN, 0, 10, False, 0),
            (2, 4, np.NaN, 0, 10, False, 0),
        )
        self.sa_allclose(
            UncertaintyBase.from_tuples(*values),
            self.get_right_2d_array(),
        )

    def sa_allclose(self, a, b):
        """allclose for structured arrays"""
        for name in a.dtype.names:
            print name, a[name], b[name]
            nan_mask_a = np.isnan(a[name])
            nan_mask_b = np.isnan(b[name])
            self.assertTrue(np.allclose(nan_mask_a, nan_mask_b))
            self.assertTrue(np.allclose(
                a[name][~nan_mask_a],
                b[name][~nan_mask_b]
            ))
        return True



from .errors import MultipleRowParamsArrayError
from functools import wraps
import numpy as np

BASE_DTYPE = [
    ('loc', np.float64),
    ('scale', np.float64),
    ('shape', np.float64),
    ('minimum', np.float64),
    ('maximum', np.float64),
    ('negative', np.bool)
]


def flatten_numpy_array(obj):
    if not isinstance(obj, np.ndarray):
        return obj
    else:
        return obj.ravel()


def one_row_params_array(function):
    @wraps(function)
    def wrapper(cls, params, *args, **kwargs):
        if len(params.shape) == 1:
            params = params.reshape(params.shape[0], 1)
        else:
            if params.shape[0] != 1:
                raise MultipleRowParamsArrayError
        # Flatten any additional inputs to one dimension
        # Needed for PDF optional xs input
        args = [flatten_numpy_array(x) for x in args]
        kwargs = dict([(key, flatten_numpy_array(obj)) for key, obj in
                      kwargs.iteritems()])
        return function(cls, params, *args, **kwargs)
    return wrapper


def construct_params_array(length=1, include_type=False):
    dtype = BASE_DTYPE
    if include_type:
        dtype = dtype + [('uncertainty_type', np.uint8)]
    params = np.zeros((length,), dtype=dtype)
    params['minimum'] = params['maximum'] = np.NaN
    params['scale'] = params['loc'] = params['shape'] = np.NaN
    return params



from __future__ import division
from ...distributions import GeneralizedExtremeValueUncertainty as GEVU
from ...errors import InvalidParamsError
from ..base import UncertaintyTestCase
import numpy as np


class GeneralizedExtremeValueUncertaintyTestCase(UncertaintyTestCase):

    def test_random_variables(self):
        params = self.make_params_array()
        params['loc'] = 2
        params['scale'] = 5
        # Formula for median (loc - scale * ln ln 2)
        expected_median = 2 - 5 * np.log(np.log(2))
        results = GEVU.random_variables(params, 10000)
        found_median = np.median(results)
        self.assertEqual(results.shape, (1, 10000))
        self.assertTrue(0.95 * expected_median < found_median)
        self.assertTrue(found_median < 1.05 * expected_median)

    def test_loc_validation(self):
        params = self.make_params_array()
        params['loc'] = np.NaN
        self.assertRaises(
            InvalidParamsError,
            GEVU.validate,
            params
        )

    def test_scale_validation(self):
        params = self.make_params_array()
        params['scale'] = -1
        self.assertRaises(
            InvalidParamsError,
            GEVU.validate,
            params
        )

    def test_shape_validation(self):
        params = self.make_params_array()
        params['shape'] = 1
        self.assertRaises(
            InvalidParamsError,
            GEVU.validate,
            params
        )

    def make_params_array(self, length=1):
        assert isinstance(length, int)
        params = np.zeros((length,), dtype=[
            ('input', 'u4'),
            ('output', 'u4'),
            ('loc', 'f4'),
            ('negative', 'b1'),
            ('scale', 'f4'),
            ('shape', 'f4'),
            ('minimum', 'f4'),
            ('maximum', 'f4')
        ])
        params['minimum'] = params['maximum'] = np.NaN
        params['loc'] = params['scale'] = 1
        return params



from __future__ import division
from .base import BoundedUncertaintyBase
from numpy import random, zeros, array


class BernoulliUncertainty(BoundedUncertaintyBase):
    id = 6
    description = "Bernoulli uncertainty"

    @classmethod
    def random_variables(cls, params, size, seeded_random=None):
        data = zeros((params.shape[0], size))
        if not seeded_random:
            seeded_random = random
        adjusted_means, scale = cls.rescale(params)
        adjusted_means.resize(params.shape[0], 1)
        # nums = seeded_random.random_sample(size * params.shape[0])
        # nums_rs = nums.reshape(params.shape[0], size)
        mask = seeded_random.random_sample(size * params.shape[0]).reshape(
            (params.shape[0], size)) >= adjusted_means
        data[mask] = 1
        return data

    @classmethod
    def cdf(cls, params, vector):
        vector = cls.check_2d_inputs(params, vector)
        # Turn into column vector
        mean = array(params['loc'])
        mean.resize(params.shape[0], 1)
        return (vector >= mean) * 1

    @classmethod
    def ppf(cls, params, percentages):
        percentages = cls.check_2d_inputs(params, percentages)
        adjusted_means, scale = cls.rescale(params)
        length = params.shape[0]
        return (
            (percentages >= adjusted_means.reshape(length, 1))
            * scale.reshape(length, 1)
            + params['minimum'].reshape(length, 1)
        )



# -*- coding: utf-8 -*
from ..errors import InvalidParamsError,\
    ImproperBoundsError
from ..utils import one_row_params_array
from .base import UncertaintyBase
from scipy import stats
import numpy as np


class DiscreteUniform(UncertaintyBase):

    """
The discrete uniform distribution includes all integer values from the ``minimum`` up to, but excluding the ``maximum``.

See https://en.wikipedia.org/wiki/Uniform_distribution_(discrete).
    """
    id = 7
    description = "Discrete uniform uncertainty"

    @classmethod
    def validate(cls, params):
        # No mean value
        if np.isnan(params['maximum']).sum():
            raise InvalidParamsError("Maximum values must always be defined.")
        # Minimum <= Maximum
        if (params['minimum'] >= params['maximum']).sum():
            raise ImproperBoundsError

    @classmethod
    def fix_nan_minimum(cls, params):
        mask = np.isnan(params['minimum'])
        params['minimum'][mask] = 0
        return params

    @classmethod
    def random_variables(cls, params, size, seeded_random=None):
        if not seeded_random:
            seeded_random = np.random
        params = cls.fix_nan_minimum(params)
        # randint has different behaviour than e.g. uniform. We can't pass in
        # arrays, but have to process them line by line.
        return np.vstack([
            seeded_random.randint(
                params['minimum'][i],  # Minimum (low)
                params['maximum'][i],  # Maximum (high)
                size=size
            ) for i in range(params.shape[0])
        ])

    @classmethod
    def cdf(cls, params, vector):
        vector = cls.check_2d_inputs(params, vector)
        results = np.zeros(vector.shape)
        params = cls.fix_nan_minimum(params)
        for row in range(params.shape[0]):
            results[row, :] = stats.randint.cdf(
                vector[row, :],
                loc=params[row]['minimum'],
                scale=params[row]['maximum'] - params[row]['minimum']
            )
        return results

    @classmethod
    def ppf(cls, params, percentages):
        percentages = cls.check_2d_inputs(params, percentages)
        params = cls.fix_nan_minimum(params)
        scale = (params['maximum'] - params['minimum']).reshape(
            params.shape[0], 1)
        return percentages * scale + params['minimum'].reshape(
            params.shape[0], 1)

    @classmethod
    @one_row_params_array
    def statistics(cls, params):
        params = cls.fix_nan_minimum(params)
        mean = (params['maximum'] + params['minimum']) / 2
        return {
            'mean': mean,
            'mode': "Undefined",
            'median': int(round(mean)),
            'lower': params['minimum'],
            'upper': params['maximum']
        }

    @classmethod
    @one_row_params_array
    def pdf(cls, params, xs=None):
        params = cls.fix_nan_minimum(params)
        if xs is None:
            xs = (params['minimum'], params['maximum'])
        percentage = 1 / (params['maximum'] - params['minimum'])
        ys = np.array([float(percentage) for x in xs])
        return np.array([float(x) for x in xs]), ys



from __future__ import division
from .base import UncertaintyBase
from ..errors import InvalidParamsError
import numpy as np


class GammaUncertainty(UncertaintyBase):

    r"""
The Gamma uncertainty distribution probability density function as a function of :math:`k`, the shape parameters, and :math:`\theta`, the scale parameter:

.. math:: f(x;k,\theta) =  \frac{x^{k-1}e^{-\frac{x}{\theta}}}{\theta^k\Gamma(k)}

The scale parameter :math:`k` is ``shape``, and :math:`\theta` is ``scale``. An optional location parameter, which offsets the distribution from the origin, can be specified in ``loc``.

See https://en.wikipedia.org/wiki/Gamma_distribution.
    """

    id = 9
    description = "Gamma uncertainty"

    @classmethod
    def validate(cls, params, transform=False):
        if (params['shape'] <= 0).sum() or np.isnan(params['shape']).sum():
            raise InvalidParamsError(
                "Positive shape (k) values required for Gamma distribution."
            )
        if (params['scale'] <= 0).sum() or np.isnan(params['scale']).sum():
            raise InvalidParamsError(
                "Positive scale (theta) values required for Gamma distribution."
            )

    @classmethod
    def random_variables(cls, params, size, seeded_random=None,
                         **kwargs):
        if seeded_random is None:
            seeded_random = np.random
        offset = params['loc'].copy()
        offset[np.isnan(offset)] = 0
        data = offset.reshape((-1, 1)) + seeded_random.gamma(
            shape=params['shape'],
            scale=params['scale'],
            size=(size, params.shape[0])).T
        data[params['negative'], :] = -1 * data[params['negative'], :]
        return data



from __future__ import division
from ..errors import InvalidParamsError
from ..utils import one_row_params_array
from .base import UncertaintyBase
from scipy import stats
import numpy as np


class LognormalUncertainty(UncertaintyBase):
    id = 2
    description = "Lognormal uncertainty"

    @classmethod
    def validate(cls, params):
        """Custom validation because mean gets log-transformed"""
        if np.isnan(params['loc']).sum():
            raise InvalidParamsError(
                "Real location (mu) values are required for"
                " lognormal uncertainties."
            )
        if np.isnan(params['scale']).sum() or (params['scale'] <= 0).sum():
            raise InvalidParamsError(
                "Real, positive scale (sigma) values are required for"
                " lognormal uncertainties."
            )

    @classmethod
    def random_variables(cls, params, size, seeded_random=None):
        if not seeded_random:
            seeded_random = np.random
        data = seeded_random.lognormal(
            params['loc'],  # Mu
            params['scale'],  # Sigma
            size=(size, params.shape[0])
        ).T
        # Negative is needed because log loses sign information.
        # Error handling not included, as this loop is called many times
        data[params['negative'], :] = -1 * data[params['negative'], :]
        return data

    @classmethod
    def cdf(cls, params, vector):
        vector = cls.check_2d_inputs(params, vector)
        vector[params['negative']] = -1 * vector[params['negative']]
        results = np.zeros(vector.shape)
        for row in range(params.shape[0]):
            results[row, :] = stats.lognorm.cdf(
                vector[row, :],
                params['scale'][row],
                scale=np.exp(params['loc'][row])
            )
        return results

    @classmethod
    def ppf(cls, params, percentages):
        percentages = cls.check_2d_inputs(params, percentages)
        results = np.zeros(percentages.shape)
        for row in range(percentages.shape[0]):
            results[row, :] = stats.lognorm.ppf(
                percentages[row, :],
                params['scale'][row],
                scale=np.exp(params['loc'][row])
            )
        results[params['negative']] = -1 * results[params['negative']]
        return results

    @classmethod
    @one_row_params_array
    def statistics(cls, params):
        negative = -1 if bool(params['negative']) else 1
        geometric_mu = float(np.exp(params['loc']))
        sigma = float(params['scale'])
        mu = float(params['loc'])
        geometric_sigma = float(np.exp(sigma))
        mean = np.exp(mu + (sigma ** 2) / 2)
        mode = np.exp(mu - sigma ** 2)
        ci_95_lower = geometric_mu / (geometric_sigma ** 2)
        ci_95_upper = geometric_mu * (geometric_sigma ** 2)
        if negative == -1:
            ci_95_lower, ci_95_upper = ci_95_upper, ci_95_lower
        return {
            'median': negative * geometric_mu,
            'mode': negative * mode,
            'mean': negative * mean,
            'lower': negative * ci_95_lower,
            'upper': negative * ci_95_upper
        }

    @classmethod
    @one_row_params_array
    def pdf(cls, params, xs=None):
        """Generate probability distribution function for lognormal distribution."""
        n = params['negative']
        if xs is None:
            # Find nice range to graph
            lower = ((-1 if n else 1) *
                       params['loc'] / (np.exp(params['scale']) **
                       cls.standard_deviations_in_default_range))[0, 0]
            upper = ((-1 if n else 1) *
                       params['loc'] * (np.exp(params['scale']) **
                       cls.standard_deviations_in_default_range))[0, 0]
            if n:
                lower, upper = upper, lower

            xs = np.linspace(
                (lower if np.isnan(params['minimum']) else params['minimum']).ravel(),
                (upper if np.isnan(params['maximum']) else params['maximum']).ravel(),
                cls.default_number_points_in_pdf
            ).ravel()

        ys = stats.lognorm.pdf(
            -1 * xs if n else xs,
            params['scale'],
            scale=np.exp(params['loc'])
        )
        return xs, ys.ravel()



# -*- coding: utf-8 -*
from __future__ import division
from .errors import UnknownUncertaintyType
from .uncertainty_choices import *
import numpy as np


class RandomNumberGenerator(object):

    def __init__(self, uncertainty_type, params, size=1,
                 maximum_iterations=100, seed=None,
                 **kwargs):
        """
Create a random number generator from a :ref:`params-array` and an uncertainty distribution.

Upon instantiation, the class checks that:

* The ``minimum`` and ``maximum`` bounds, if any, are reasonable
* The given uncertainty type can be used

``uncertainty_type`` is not required to be a subclass of :ref:`UncertaintyBase`, but needs to have the method ``bounded_random_variables``.

The returned class instance can be called directly::

    >>> from stats_arrays import RandomNumberGenerator, TriangularUncertainty
    >>> params = TriangularUncertainty.from_dicts(
    ...     {'loc': 5, 'minimum': 3, 'maximum': 10},
    ...     {'loc': 1, 'minimum': 0.7, 'maximum': 4.4}
    ...     )
    >>> rng = RandomNumberGenerator(TriangularUncertainty, params)
    >>> rng.generate_random_numbers()
    array([[ 8.00843856],
       [ 1.54968237]])

but can also be used as an iterator::

    >>> zip(range(2), rng)
    [(0, array([[ 5.34298156],
       [ 1.02447677]])),
     (1, array([[ 5.45360508],
       [ 1.99372889]]))]

Args:
    * **uncertainty_type** (object): An uncertainty type class (subclass of ``stats_arrays.distributions.UncertaintyBase``)
    * **params** (array): The :ref:`params-array`
    * *size* (int, optional): The number of samples to draw from each parameter. Default is ``1``.
    * *maximum_iterations* (int, optional): The number of times to draw samples that fit within the given bounds, if any, before raising ``stats_arrays.MaximumIterationsError``. Default is ``100``.
    * *seed* (int, optional): Seed value for the random number generator. Default is ``None``.

Returns:
    A class instance

        """
        self.params = params
        self.length = self.params.shape[0]
        self.size = size
        self.uncertainty_type = uncertainty_type
        self.maximum_iterations = maximum_iterations
        self.random = np.random.RandomState(seed)
        self.verify_uncertainty_type()
        self.verify_params()

    def verify_params(self, params=None, uncertainty_type=None):
        """Verify that parameters are within bounds. Mean is not restricted to bounds, unless the distribution requires it (e.g. triangular)."""
        if params is None:  # Can't convert array to boolean
            params = self.params
        if not uncertainty_type:
            uncertainty_type = self.uncertainty_type
        uncertainty_type.validate(params)

    def verify_uncertainty_type(self, uncertainty_type=None):
        """Make sure the given uncertainty type provides the method ``bounded_random_variables``."""
        if not uncertainty_type:
            uncertainty_type = self.uncertainty_type
        if not hasattr(uncertainty_type, "bounded_random_variables"):
            raise UnknownUncertaintyType(
                "The provided uncertainty type must have the "
                "`bounded_random_variables` method."
            )

    def generate_random_numbers(self, uncertainty_type=None, params=None,
                                size=None):
        if not uncertainty_type:
            uncertainty_type = self.uncertainty_type
        if params is None:  # Can't convert array to boolean
            params = self.params
        if not size:
            size = self.size
        return uncertainty_type.bounded_random_variables(
            params,
            size,
            self.random,
            self.maximum_iterations
        )

    def next(self):
        return self.generate_random_numbers()

    def __iter__(self):
        return self


class MCRandomNumberGenerator(object):

    """
A Monte Carlo random number generator that operates on a :ref:`hpa`.

Upon instantiation, the class checks that:

* Each unique ``uncertainty_type`` is a valid choice in ``uncertainty_choices``
* That the parameter array for each uncertainty type validates

The returned class instance can be called directly with ``next``, or can be used as an iterator::

    >>> from stats_arrays import MCRandomNumberGenerator, UncertaintyBase
    >>> params = UncertaintyBase.from_dicts(
    ...     {'loc': 5, 'minimum': 3, 'maximum': 10, 'uncertainty_type': 5},
    ...     {'loc': 1, 'scale': 0.7, 'uncertainty_type': 3}
    ...     )
    >>> mcrng = MCRandomNumberGenerator(params)
    >>> zip(range(2), mcrng)
    [(0, array([ 1.35034874,  5.2705415 ])),
     (1, array([ 5.2705415 ,  1.35034874]))]

Args:
    * **params** (array): The :ref:`hpa`
    * *maximum_iterations* (int, optional): The number of times to draw samples that fit within the given bounds, if any, before raising ``stats_arrays.MaximumIterationsError``. Default is ``100``.
    * *seed* (int, optional): Seed value for the random number generator. Default is ``None``.

Returns:
    A class instance

    """

    def __init__(self, params, maximum_iterations=50, seed=None, **kwargs):
        self.params = params.copy()
        self.length = self.params.shape[0]
        self.maximum_iterations = maximum_iterations
        self.choices = uncertainty_choices
        self.random = np.random.RandomState(seed)
        self.verify_params()
        self.ordering = np.argsort(self.params["uncertainty_type"])
        self.params = self.params[self.ordering]
        self.positions = self.get_positions()

    def get_positions(self):
        """Construct dictionary of where each distribution starts and stops in the sorted parameter array"""
        return dict([(choice, (self.params['uncertainty_type'] == choice.id).sum()
                      ) for choice in self.choices])

    def verify_params(self):
        """Verify that all uncertainty types are allowed, and parameter validate using distribution class methods"""
        ids = set(np.unique(self.params['uncertainty_type']))
        extra_ids = ids.difference(set([x.id for x in self.choices]))
        if extra_ids:
            raise ValueError(
                "Uncertainty type id(s) %s are not valid" % extra_ids
            )

        for uncertainty_type in self.choices:
            mask = self.params['uncertainty_type'] == uncertainty_type.id
            if mask.sum():
                uncertainty_type.validate(self.params[mask])

    def next(self):
        """Generate a new vector of random numbers"""
        if not hasattr(self, "random_data"):
            self.random_data = np.zeros(self.length)

        offset = 0
        for uncertainty_type in self.choices:
            size = self.positions[uncertainty_type]
            if not size:
                continue
            random_data = uncertainty_type.bounded_random_variables(
                self.params[offset:size + offset],
                1,
                self.random,
                self.maximum_iterations
            )
            if len(random_data.shape) == 2:
                random_data = random_data[:, 0]  # Restore to 1-d
            self.random_data[offset:size + offset] = random_data
            offset += size

        self.random_data = self.random_data[np.argsort(self.ordering)]
        return self.random_data

    def __iter__(self):
        return self


class LatinHypercubeRNG(MCRandomNumberGenerator):

    """
A random number generator that pre-calculates a sample space to draw from.

.. rubric:: Inputs

* params : A :ref:`params-array` which gives parameters for distributions (one distribution per row).
* seed : An integer (or array of integers) to seed the `NumPy random number generator <http://docs.scipy.org/doc/numpy/reference/generated/numpy.random.mtrand.RandomState.html#numpy.random.mtrand.RandomState>`_.
* samples : An integer number of samples to construct for each distribution.
    """

    def __init__(self, params, seed=None, samples=10, **kwargs):
        self.params = params
        self.length = self.params.shape[0]
        self.row_index = np.arange(self.length)
        self.size = 1
        self.samples = samples
        self.choices = uncertainty_choices
        self.random = np.random.RandomState(seed)
        self.uncertainty_type = None

        self.verify_params()
        self.build_hypercube()

    def build_hypercube(self):
        """Build an array, of shape `self.length` rows by `self.samples` columns, which contains the sample space to be drawn from when doing Latin Hypercubic sampling.

Each row represents a different data point and distribution. The final sample space is `self.hypercube`. All distributions from `uncertainty_choices` are usable, and bounded distributions are also fine.

.. rubric:: Builds

self.hypercube : Numpy array with dimensions `self.length` by `self.samples`."""
        step_size = 1 / (self.samples + 1)
        # Define the beginning points and step sizes - not all the same because
        # some distributions are bounded. Make adjustments to generic values
        # later.
        inputs = np.tile(
            np.array((step_size, step_size)),
            self.length
        ).reshape(self.length, 2)
        for uncertainty_type in self.choices:
            mask = self.params['uncertainty_type'] == uncertainty_type.id
            if not mask.sum():
                continue
            subarray = self.params[mask]
            # Adjust inputs when bounds are present. Easiest to do in three
            # discrete steps. First, when only a lower bound is present.
            only_min_mask = ~np.isnan(subarray['minimum']) * \
                np.isnan(subarray['maximum'])
            if only_min_mask.sum():
                mins = uncertainty_type.cdf(subarray[only_min_mask],
                                            subarray[only_min_mask]['minimum'])
                steps = (1 - mins) / (self.samples + 1)
                inputs[mask, :][only_min_mask] = np.hstack((mins + steps,
                                                            steps))
            # Next, if only a max bound is present
            only_max_mask = np.isnan(subarray['minimum']) * \
                ~np.isnan(subarray['maximum'])
            if only_max_mask.sum():
                maxs = uncertainty_type.cdf(subarray[only_max_mask],
                                            subarray[only_max_mask]['maximum'])
                steps = maxs / (self.samples + 1)
                inputs[mask, :][only_max_mask] = np.hstack((steps, steps))
            # Finally, if both min and max bounds are present
            both_mask = ~np.isnan(subarray['minimum']) * \
                ~np.isnan(subarray['maximum'])
            if both_mask.sum():
                mins = uncertainty_type.cdf(subarray[both_mask],
                                            subarray[both_mask]['minimum'])
                maxs = uncertainty_type.cdf(subarray[both_mask],
                                            subarray[both_mask]['maximum'])
                steps = (maxs - mins) / (self.samples + 1)
                inputs[mask, :][both_mask] = np.hstack((mins + steps,
                                                        steps))
        # Percentages is now a list, samples wide, of the percentages covered
        # by the bounded or unbounded distributions.
        self.percentages = inputs[:, 0].reshape(self.length, 1) + np.arange(0,
                                                                            self.samples, 1) * inputs[:, 1].reshape(self.length, 1)
        # Transform percentages into a sample space
        self.hypercube = np.zeros((self.length, self.samples))
        for uncertainty_type in self.choices:
            mask = self.params['uncertainty_type'] == uncertainty_type.id
            if not mask.sum():
                continue
            self.hypercube[mask, :] = \
                uncertainty_type.ppf(params=self.params[mask],
                                     percentages=self.percentages[mask, :])
        self.hypercube[self.params['negative'], :] = \
            -1 * self.hypercube[self.params['negative'], :]

    def next(self):
        """Draw directly from pre-computed sample space."""
        return self.hypercube[self.row_index,
                              self.random.randint(self.samples, size=self.length)]



import unittest
from nose.plugins.skip import SkipTest
from numpy import *
from ..random import LatinHypercubeRNG
from ..distributions import *


class LatinHypercubeMCTestCase(unittest.TestCase):

    def make_params_array(self, length=1):
        assert isinstance(length, int)
        params = zeros((length,), dtype=[('uncertainty_type', 'i2'),
                                         ('input', 'u4'), ('output', 'u4'),
                                         ('loc', 'f4'), ('negative', 'b1'), ('scale', 'f4'),
                                         ('minimum', 'f4'), ('maximum', 'f4')])
        params['minimum'] = params['maximum'] = params['scale'] = NaN
        return params

    def test_known_seed(self):
        params = self.make_params_array(1)
        params['uncertainty_type'] = UniformUncertainty.id
        params['maximum'] = 1
        params['minimum'] = 0
        params['loc'] = 0.5
        params['scale'] = NaN
        lhc = LatinHypercubeRNG(params, seed=11111, samples=9)
        self.assertTrue(allclose(lhc.iterate(), array([[0.5]])))
        self.assertTrue(allclose(lhc.iterate(), array([[0.7]])))
        self.assertTrue(allclose(lhc.iterate(), array([[0.5]])))
        self.assertTrue(allclose(lhc.iterate(), array([[0.6]])))

    def test_no_uncertainty_intervals(self):
        # Two-dimensional array check
        params = self.make_params_array(2)
        params['uncertainty_type'] = NoUncertainty.id
        params['loc'] = (1, 2)
        lhc = LatinHypercubeRNG(params)
        test_values = array(([1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                             [2, 2, 2, 2, 2, 2, 2, 2, 2, 2]))
        self.assertTrue(allclose(lhc.hypercube, test_values))
        params['uncertainty_type'] = UndefinedUncertainty.id
        lhc = LatinHypercubeRNG(params)
        self.assertTrue(allclose(lhc.hypercube, test_values))
        # One-dimensional array check
        params = self.make_params_array(1)
        params['loc'] = 1
        params['uncertainty_type'] = NoUncertainty.id
        lhc = LatinHypercubeRNG(params)
        test_values = array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1]).reshape(1, 10)
        self.assertTrue(allclose(lhc.hypercube, test_values))

    @SkipTest
    def test_lognormal_intervals(self):
        # Two-dimensional array check
        params = self.make_params_array(2)
        params['uncertainty_type'] = LognormalUncertainty.id
        params['loc'] = (1, 2)
        params['scale'] = (1, 2)
        lhc = LatinHypercubeRNG(params)
        test_values = array([[0.26311141, 0.40314545, 0.54630091, 0.70556548,
                              0.89209264, 1.12095981, 1.41730289, 1.83049303, 2.48049433, 3.8006714],
                             [0.52622281, 0.8062909, 1.09260181, 1.41113097, 1.78418529, 2.24191963,
                              2.83460579, 3.66098607, 4.96098868, 7.60134282]])
        # print lhc.hypercube
        # print test_values
        self.assertTrue(allclose(lhc.hypercube, test_values))
        # One-dimensional array check
        params = self.make_params_array(1)
        params['loc'] = 1
        params['scale'] = 1
        params['uncertainty_type'] = LognormalUncertainty.id
        lhc = LatinHypercubeRNG(params)
        test_values = array([0.26311141, 0.40314545, 0.54630091, 0.70556548,
                             0.89209264, 1.12095981, 1.41730289, 1.83049303, 2.48049433, 3.8006714]
                            ).reshape(1, 10)
        self.assertTrue(allclose(lhc.hypercube, test_values))
        # Lower bound check
        params['loc'] = 1
        params['minimum'] = 1
        lhc = LatinHypercubeRNG(params)
        test_values = array([[1.12095981, 1.25845417, 1.41730289, 1.604463,
                              1.83049303, 2.11247151, 2.48049433, 2.99457872, 3.8006714, 5.42285066]
                             ]).reshape(1, 10)
        self.assertTrue(allclose(lhc.hypercube, test_values))
        # Upper bound check
        params['loc'] = 1
        params['maximum'] = 1
        params['minimum'] = NaN
        lhc = LatinHypercubeRNG(params)
        test_values = array([[0.18440486, 0.26311141, 0.33393679, 0.40314545,
                              0.47337916, 0.54630091, 0.62326149, 0.70556548, 0.79462568,
                              0.89209264]]).reshape(1, 10)
        self.assertTrue(allclose(lhc.hypercube, test_values))
        # Both bounds check
        params['loc'] = 1
        params['minimum'] = 0.5
        params['maximum'] = 1.5
        lhc = LatinHypercubeRNG(params)
        test_values = array([[0.56110473, 0.62514369, 0.69285366, 0.76502388,
                              0.84254518, 0.92646272, 1.01804071, 1.11884934, 1.23088804, 1.35676875]
                             ]).reshape(1, 10)
        self.assertTrue(allclose(lhc.hypercube, test_values))

    @SkipTest
    def test_normal_intervals(self):
        # Two-dimensional array check
        params = self.make_params_array(2)
        params['uncertainty_type'] = NormalUncertainty.id
        params['loc'] = (1, 2)
        params['scale'] = (1, 2)
        lhc = LatinHypercubeRNG(params)
        test_values = array(([-0.33517774,  0.09154213,  0.39541465,  0.6512443,
                              0.88581471, 1.11418529, 1.3487557, 1.60458535, 1.90845787,
                              2.33517774], [-0.67035547,  0.18308426,  0.79082931,  1.30248861,
                                            1.77162941,  2.22837059,  2.69751139,  3.20917069,  3.81691574,
                                            4.67035547]))
        self.assertTrue(allclose(lhc.hypercube, test_values))
        # One-dimensional array check
        params = self.make_params_array(1)
        params['loc'] = 1
        params['scale'] = 1
        params['uncertainty_type'] = NormalUncertainty.id
        lhc = LatinHypercubeRNG(params)
        test_values = array([-0.33517774,  0.09154213,  0.39541465,
                             0.6512443, 0.88581471,  1.11418529,  1.3487557,  1.60458535,
                             1.90845787, 2.33517774]).reshape(1, 10)
        self.assertTrue(allclose(lhc.hypercube, test_values))
        # Lower bound check
        params['minimum'] = 1
        lhc = LatinHypercubeRNG(params)
        test_values = array([[1.11418529, 1.22988412, 1.3487557, 1.47278912,
                              1.60458535, 1.74785859, 1.90845787, 2.09680356, 2.33517774,
                              2.69062163]]).reshape(1, 10)
        self.assertTrue(allclose(lhc.hypercube, test_values))
        # Upper bound check
        params['maximum'] = 1
        params['minimum'] = NaN
        lhc = LatinHypercubeRNG(params)
        test_values = array([[-0.69062163, -0.33517774, -0.09680356,
                              0.09154213, 0.25214141, 0.39541465, 0.52721088, 0.6512443,
                              0.77011588, 0.88581471]]).reshape(1, 10)
        self.assertTrue(allclose(lhc.hypercube, test_values))
        # Both bounds check
        params['minimum'] = 0.5
        params['maximum'] = 1.5
        lhc = LatinHypercubeRNG(params)
        test_values = array([[0.59665956, 0.68968379, 0.78009253, 0.86873532,
                              0.95635658, 1.04364342, 1.13126468, 1.21990747, 1.31031621,
                              1.40334044]]).reshape(1, 10)
        self.assertTrue(allclose(lhc.hypercube, test_values))

    def test_uniform_intervals(self):
        # Two-dimensional array check
        params = self.make_params_array(2)
        params['uncertainty_type'] = UniformUncertainty.id
        params['maximum'] = 2
        params['minimum'] = 1
        params['loc'] = 1.5
        params['scale'] = NaN
        lhc = LatinHypercubeRNG(params, samples=9)
        self.assertTrue(allclose(lhc.hypercube, array([[1.1, 1.2, 1.3, 1.4,
                                                        1.5, 1.6, 1.7, 1.8, 1.9], [1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8,
                                                                                   1.9]])))
        # One-dimensional array check
        params = self.make_params_array(1)
        params['uncertainty_type'] = UniformUncertainty.id
        params['maximum'] = 2
        params['minimum'] = 1
        params['loc'] = 1.5
        params['scale'] = NaN
        lhc = LatinHypercubeRNG(params, samples=9)
        self.assertTrue(allclose(lhc.hypercube,
                                 array([[1.1, 1.2, 1.3, 1.4, 1.5,
                                         1.6, 1.7, 1.8, 1.9]])))

    def test_triangular_intervals(self):
        # Two-dimensional array check
        params = self.make_params_array(2)
        params['uncertainty_type'] = TriangularUncertainty.id
        params['maximum'] = 2
        params['minimum'] = 1
        params['loc'] = 1.6
        params['scale'] = NaN
        lhc = LatinHypercubeRNG(params, samples=9)
        self.assertTrue(allclose(lhc.hypercube, array([[1.24494898, 1.34641017,
                                                        1.42426408, 1.48989796, 1.54772257, 1.60000001, 1.65358985, 1.7171573,
                                                        1.80000001], [1.24494898, 1.34641017, 1.42426408, 1.48989796,
                                                                      1.54772257, 1.60000001, 1.65358985, 1.7171573, 1.80000001]])))
        # One-dimensional array check
        params = self.make_params_array(1)
        params['uncertainty_type'] = TriangularUncertainty.id
        params['maximum'] = 2
        params['minimum'] = 1
        params['loc'] = 1.6
        params['scale'] = NaN
        lhc = LatinHypercubeRNG(params, samples=9)
        self.assertTrue(allclose(lhc.hypercube, array([[1.24494898, 1.34641017,
                                                        1.42426408, 1.48989796, 1.54772257, 1.60000001, 1.65358985, 1.7171573,
                                                        1.80000001]])))

    def test_bernoulli(self):
        # Two-dimensional array check
        params = self.make_params_array(2)
        params['uncertainty_type'] = BernoulliUncertainty.id
        params['maximum'] = 2
        params['minimum'] = 1
        params['loc'] = 1.6
        lhc = LatinHypercubeRNG(params)
        self.assertTrue(allclose(lhc.hypercube, array([[1, 1, 1, 1, 1, 1, 2, 2, 2, 2],
                                                       [1, 1, 1, 1, 1, 1, 2, 2, 2, 2]])))
        # One-dimensional array check
        params = self.make_params_array(1)
        params['uncertainty_type'] = BernoulliUncertainty.id
        params['maximum'] = 2
        params['minimum'] = 1
        params['loc'] = 1.6
        lhc = LatinHypercubeRNG(params)
        self.assertTrue(allclose(lhc.hypercube, array([1, 1, 1, 1, 1, 1, 2, 2, 2, 2])))

    def test_heterogeneous_params(self):
        params = self.make_params_array(2)
        params['uncertainty_type'][0] = TriangularUncertainty.id
        params['uncertainty_type'][1] = UniformUncertainty.id
        params['maximum'] = 2
        params['minimum'] = 1
        params['loc'] = 1.6
        params['scale'] = NaN
        lhc = LatinHypercubeRNG(params, samples=9)
        self.assertTrue(allclose(lhc.hypercube, array([[1.24494898, 1.34641017,
                                                        1.42426408, 1.48989796, 1.54772257, 1.60000001, 1.65358985, 1.7171573,
                                                        1.80000001], [1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9]])))



from __future__ import division
from ..errors import UndefinedDistributionError
from .base import UncertaintyBase
from numpy import repeat, tile


class UndefinedUncertainty(UncertaintyBase):

    """Undefined or unknown uncertainty"""
    id = 0
    description = "Undefined or unknown uncertainty"

    @classmethod
    def random_variables(cls, params, size, seeded_random=None):
        return repeat(params['loc'], size).reshape((params.shape[0],
                                                    size))

    @classmethod
    def cdf(cls, params, vector):
        raise UndefinedDistributionError(
            "Can't calculate percentages for an undefined distribution.")

    @classmethod
    def ppf(cls, params, percentages):
        return tile(params['loc'].reshape((params.shape[0], 1)),
                    percentages.shape[1])


class NoUncertainty(UndefinedUncertainty):
    id = 1
    description = "No uncertainty"



from ...errors import InvalidParamsError
from ...distributions import LognormalUncertainty as LU
from ..base import UncertaintyTestCase
from scipy.special import erf
import numpy as np


class LognormalTestCase(UncertaintyTestCase):

    def pdf(self, x, mu, sigma):
        return 1 / (x * np.sqrt(2 * np.pi * sigma ** 2)
                    ) * np.e ** (-((np.log(x) - mu) ** 2) / (
            2 * sigma ** 2))

    def cdf(self, x, mu, sigma):
        return 0.5 * (1 + erf((np.log(x) - mu) / np.sqrt(
            2 * sigma ** 2)))

    def test_pdf_positive(self):
        mu, sigma = np.random.random() / 5 + 0.2, np.random.random() / 10 + 0.1
        pa = LU.from_dicts({'loc': mu, 'scale': sigma})
        self.assertTrue(np.allclose(
            self.pdf(1.2, mu, sigma),
            # [0] are X values, [1] are Y values
            LU.pdf(pa, np.array((1.2,)))[1]
        ))

    def test_pdf_negative(self):
        mu, sigma = np.random.random() / 5 + 0.2, np.random.random() / 10 + 0.1
        pa = LU.from_dicts({'loc': mu, 'scale': sigma, 'negative': True})

        self.assertTrue(np.allclose(
            self.pdf(1.2, mu, sigma),
            # [0] are X values, [1] are Y values
            LU.pdf(pa, np.array((-1.2,)))[1]
        ))

    def test_pdf_bounds(self):
        pa = LU.from_dicts({'loc': 1, 'scale': 0.5, 'minimum': 1, 'maximum': 2})
        xs, ys = LU.pdf(pa)
        print xs.min(), xs.max()
        self.assertEqual(xs.min(), 1)
        self.assertEqual(xs.max(), 2)

        pa = LU.from_dicts({'loc': 1, 'scale': 0.5, 'minimum': 1})
        xs, ys = LU.pdf(pa)
        self.assertEqual(xs.min(), 1)
        self.assertTrue(xs.max() > 2)

        pa = LU.from_dicts({'loc': 1, 'scale': 0.5, 'maximum': 2})
        xs, ys = LU.pdf(pa)
        self.assertTrue(xs.min() < 1)
        self.assertEqual(xs.max(), 2)

    def test_pdf_bounds_negative(self):
        pa = LU.from_dicts({'loc': 1, 'scale': 0.5, 'minimum': -2, 'maximum': -1,
                            'negative': True})
        xs, ys = LU.pdf(pa)
        self.assertEqual(xs.min(), -2)
        self.assertEqual(xs.max(), -1)

        pa = LU.from_dicts({'loc': 1, 'scale': 0.5, 'minimum': -2, 'negative': True})
        xs, ys = LU.pdf(pa)
        print xs.min(), xs.max()
        self.assertEqual(xs.min(), -2)
        self.assertTrue(xs.max() > -1)

        pa = LU.from_dicts({'loc': 1, 'scale': 0.5, 'maximum': -1, 'negative': True})
        xs, ys = LU.pdf(pa)
        self.assertTrue(xs.min() < -2)
        self.assertEqual(xs.max(), -1)

    def test_cdf_positive(self):
        mu, sigma = np.random.random() / 5 + 0.2, np.random.random() / 10 + 0.1
        pa = LU.from_dicts({'loc': mu, 'scale': sigma})
        self.assertTrue(np.allclose(
            self.cdf(1.2, mu, sigma),
            LU.cdf(pa, np.array((1.2,)))[0]
        ))

    def test_cdf_negative(self):
        mu, sigma = np.random.random() / 5 + 0.2, np.random.random() / 10 + 0.1
        pa = LU.from_dicts({'loc': mu, 'scale': sigma, 'negative': True})
        self.assertTrue(np.allclose(
            self.cdf(1.2, mu, sigma),
            LU.cdf(pa, np.array((-1.2,)))[0]
        ))

    def test_cdf_multirow(self):
        pa = LU.from_dicts(
            {'loc': 0.4, 'scale': 0.1, 'negative': True},
            {'loc': 0.6, 'scale': 0.2, 'negative': False}
        )
        self.assertTrue(np.allclose(
            self.cdf(
                np.array((1.2, 1.5)),
                np.array((0.4, 0.6)),
                np.array((0.1, 0.2))
            ),
            LU.cdf(pa, np.array((-1.2, 1.5))).ravel()
        ))

    def test_ppf_positive(self):
        mu, sigma = np.random.random() / 5 + 0.2, np.random.random() / 10 + 0.1
        pa = LU.from_dicts({'loc': mu, 'scale': sigma})
        cdf = self.cdf(1.3, mu, sigma)
        self.assertTrue(np.allclose(
            1.3,
            LU.ppf(pa, np.array((cdf,)))[0]
        ))

    def test_ppf_negative(self):
        mu, sigma = np.random.random() / 5 + 0.2, np.random.random() / 10 + 0.1
        pa = LU.from_dicts({'loc': mu, 'scale': sigma, 'negative': True})
        cdf = self.cdf(1.3, mu, sigma)
        self.assertTrue(np.allclose(
            -1.3,
            LU.ppf(pa, np.array((cdf,)))[0]
        ))

    def test_validation(self):
        dicts = [
            {'loc': np.NaN, 'scale': 0.1},
            {'loc': 0.1, 'scale': np.NaN},
        ]
        for d in dicts:
            with self.assertRaises(InvalidParamsError):
                LU.validate(LU.from_dicts(d))

    def test_seeded_random(self):
        mu, sigma = np.random.random() / 5 + 0.2, np.random.random() / 10 + 0.1
        pa = LU.from_dicts({'loc': mu, 'scale': sigma})
        self.assertTrue(np.allclose(
            LU.random_variables(pa, 100, self.seeded_random()),
            LU.random_variables(pa, 100, self.seeded_random())
        ))

    def test_rng(self):
        mu, sigma = np.random.random() / 5 + 0.2, np.random.random() / 10 + 0.1
        median = np.exp(mu)
        ci_upper_975 = np.exp(mu) * (np.exp(sigma) ** 1.96)
        pa = LU.from_dicts({'loc': mu, 'scale': sigma})
        sample = LU.random_variables(pa, size=1e5).ravel()
        sample.sort()
        self.assertTrue(np.allclose(np.median(sample), median, 0.01))
        self.assertTrue(np.allclose(sample[97500], ci_upper_975, 0.01))

    def test_rng_negative(self):
        mu, sigma = np.random.random() / 5 + 0.2, np.random.random() / 10 + 0.1
        pa = LU.from_dicts({'loc': mu, 'scale': sigma, 'negative': True})
        sample = LU.random_variables(pa, size=100).ravel()
        self.assertEqual((sample < 0).sum(), 100)



from __future__ import division
from ...distributions import WeibullUncertainty
from ...errors import InvalidParamsError
from ..base import UncertaintyTestCase
import numpy as np


class WeibullTestCase(UncertaintyTestCase):

    def pretty_close(self, a, b):
        print a, b
        if b == 0:
            self.assertTrue(a - 0.05 < b < a + 0.05)
        else:
            self.assertTrue(0.95 * a < b < 1.05 * a)

    def test_random_variables(self):
        params = self.make_params_array()
        params['scale'] = 2  # lambda
        params['shape'] = 5  # k
        sample = WeibullUncertainty.random_variables(params, 10000)
        # Median: lambda * ln(2)^(1/k)
        self.pretty_close(2 * np.log(2) ** (1 / 5), np.median(sample))

    def test_random_variables_2d(self):
        params = self.make_params_array(2)
        params['scale'] = (5, 10)
        params['shape'] = (2, 3)
        sample = WeibullUncertainty.random_variables(params, 10000)
        self.pretty_close(5 * np.log(2) ** (1 / 2), np.median(sample[0, :]))
        self.pretty_close(10 * np.log(2) ** (1 / 3), np.median(sample[1, :]))

    def test_random_variables_offset(self):
        params = self.make_params_array(2)
        params['scale'] = (5, 10)
        params['shape'] = (2, 3)
        params['loc'] = (100, np.NaN)
        sample = WeibullUncertainty.random_variables(params, 10000)
        self.pretty_close(100 + 5 * np.log(2) ** (1 / 2), np.median(sample[0, :]))
        self.pretty_close(10 * np.log(2) ** (1 / 3), np.median(sample[1, :]))

    def test_loc_nan_ok(self):
        params = self.make_params_array()
        params['loc'] = np.NaN
        params['scale'] = 1
        params['shape'] = 1
        WeibullUncertainty.validate(params)
        return True

    def test_scale_validation(self):
        params = self.make_params_array()
        params['loc'] = np.NaN
        params['scale'] = -1
        params['shape'] = 1
        self.assertRaises(
            InvalidParamsError,
            WeibullUncertainty.validate,
            params
        )

    def test_shape_validation(self):
        params = self.make_params_array()
        params['loc'] = np.NaN
        params['scale'] = 1
        params['shape'] = -1
        self.assertRaises(
            InvalidParamsError,
            WeibullUncertainty.validate,
            params
        )



from setuptools import setup


setup(
    name='stats_arrays',
    version='0.3.1',
    author='Chris Mutel',
    author_email='cmutel@gmail.com',
    url='https://bitbucket.org/cmutel/stats_arrays',
    install_requires=["numpy", "scipy", "nose"],
    packages=[
        'stats_arrays',
        'stats_arrays.distributions',
        'stats_arrays.tests',
        'stats_arrays.tests.distributions',
    ],
    license='BSD 2-clause; LICENSE.txt',
    long_description=open('README.rst').read(),
)



from __future__ import division
from ..errors import InvalidParamsError,\
    ImproperBoundsError, UnreasonableBoundsError, \
    MaximumIterationsError
from ..utils import one_row_params_array, construct_params_array
import numpy as np

np.seterr(invalid='ignore')


class UncertaintyBase(object):

    """
Abstract base class for uncertainty types.

All methods on uncertainty classes should be `class methods <http://docs.python.org/library/functions.html#classmethod>`_, as instantiating uncertainty classes many times is not desired.

.. rubric:: Defaults

* ``default_number_points_in_pdf``: 200. The default number of points to calculate for PDF/CDF functions.
* ``standard_deviations_in_default_range``: 3. The number of standard deviations that define the default range when calculating PDF/CDF values. In a normal distribution, 3 standard deviations is approximately 99% of all values.

"""
    default_number_points_in_pdf = 200
    standard_deviations_in_default_range = 2.2

    # Conversion utilities ###
    @classmethod
    def from_tuples(cls, *data):
        """
Construct a :ref:`hpa` from parameter tuples.

The order of the parameters is:

#. ``loc``
#. ``scale``
#. ``shape``
#. ``minimum``
#. ``maximum``
#. ``negative``
#. ``uncertainty_type``

Each input tuple must have a length of exactly 7. For more flexibility, use ``from_dicts``.

Example:

.. code-block:: python

    >>> from stats_arrays import UncertaintyBase
    >>> import numpy as np
    >>> UncertaintyBase.from_tuples(
    ...     (2, 3, np.NaN, np.NaN, np.NaN, False, 3),
    ...     (5, np.NaN, np.NaN, 3, 10, False, 5)
    ...     )
    array([(2.0, 3.0, nan, nan, nan, False, 3),
           (5.0, nan, nan, 3.0, 10.0, False, 5)],
           dtype=[('loc', '<f8'), ('scale', '<f8'), ('shape', '<f8'),
                  ('minimum', '<f8'), ('maximum', '<f8'), ('negative', '?'),
                  ('uncertainty_type', 'u1')])

Args:
    One of more tuples of length 7.

Returns:
    A :ref:`hpa`

        """
        params = construct_params_array(len(data), True)
        for index, row in enumerate(data):
            params[index] = row
        return params

    @classmethod
    def from_dicts(cls, *dicts):
        """
Construct a :ref:`hpa` from parameter dictionaries.

Dictionary keys are the normal parameter array columns. Each distribution defines which columns are required and which are optional.

Example:

.. code-block:: python

    >>> from stats_arrays import UncertaintyBase
    >>> import numpy as np
    >>> UncertaintyBase.from_dicts(
    ...     {'loc': 2, 'scale': 3, 'uncertainty_type': 3},
    ...     {'loc': 5, 'minimum': 3, 'maximum': 10, 'uncertainty_type': 5}
    ...     )
    array([(2.0, 3.0, nan, nan, nan, False, 3),
           (5.0, nan, nan, 3.0, 10.0, False, 5)],
           dtype=[('loc', '<f8'), ('scale', '<f8'), ('shape', '<f8'),
                  ('minimum', '<f8'), ('maximum', '<f8'), ('negative', '?'),
                  ('uncertainty_type', 'u1')])

Args:
    One of more dictionaries.

Returns:
    A :ref:`hpa`

        """
        LABELS = [
            ('loc', np.NaN),
            ('scale', np.NaN),
            ('shape', np.NaN),
            ('minimum', np.NaN),
            ('maximum', np.NaN),
            ('negative', False),
            ('uncertainty_type', 0),
        ]
        params = construct_params_array(len(dicts), True)
        for index, obj in enumerate(dicts):
            data = [obj.get(key, default) for key, default in LABELS]
            params[index] = tuple(data)
        return params

    # Utility methods ###
    @classmethod
    def validate(cls, params):
        """
Validate the parameter array for uncertainty distribution.

Validation is distribution specific. The only default check is that ``minimum`` is less than or equal to ``maximum``, and otherwise raises ``stats_arrays.ImproperBoundsError``.

Doesn't return anything.

Args:
    A :ref:`params-array`.

    """
        # Minimum <= Maximum
        if (params['minimum'] >= params['maximum']).sum():
            raise ImproperBoundsError

    @classmethod
    def check_2d_inputs(cls, params, vector):
        """Convert ``vector`` to 2 dimensions if not already, and raise ``stats_arrays.InvalidParamsError`` if ``vector`` and ``params`` dimensions don't match."""
        if len(vector.shape) == 1:
            # Slices from structured arrays can't always be resized
            vector = np.array(vector)
            # Transform to 2-dimensional
            vector.resize(vector.shape[0], 1)
        if params.shape[0] != vector.shape[0]:
            raise InvalidParamsError(
                "Vector shape must be either (m,) or (m,n), where params has m rows. Vector has shape %s, and params is %i rows" %
                (vector.shape, params.shape[0]))
        return vector

    @classmethod
    @one_row_params_array
    def check_bounds_reasonableness(cls, params, threshold=0.1):
        """
Test if there is at least a ``threshold`` percent chance of generating random numbers within the provided bounds.

Doesn't return anything. Raises ``stats_arrays.UnreasonableBoundsError`` if this condition is not met.

.. rubric:: Inputs

* params : A one-row :ref:`params-array`.
* threshold : A percentage between 0 and 1. The minimum loc of the distribution covered by the bounds before an error is raised."""
        min_percentage = float(cls.cdf(params, params['minimum']))
        max_percentage = float(cls.cdf(params, params['maximum']))
        coverage = max_percentage - min_percentage
        if coverage < threshold:
            raise UnreasonableBoundsError(
                "The provided bounds cover only %.2f percent of the possible distribution" % coverage)

    # Used for Monte Carlo ###
    @classmethod
    def bounded_random_variables(cls, params, size, seeded_random=None,
                                 maximum_iterations=50):
        """Generate random variables repeatedly until all varaibles are within the bounds of each distribution. Raise MaximumIterationsError if this takes more that `maximum_iterations`. Uses `random_variables` for random number generation.

.. rubric:: Inputs

* params : A :ref:`params-array`.
* size : Integer. The number of values to draw from each distribution in `params`.
* seeded_random : Integer. Optional. Random seed to get repeatable samples.
* maximum_iterations : Integer. Optional. Maximum iterations to try to fit the given bounds before an error is raised.

.. rubric:: Output

An array of random values, with dimensions `params` rows by `size`."""
        data = cls.random_variables(params, size, seeded_random)
        min_array = params['minimum'].reshape(params.shape[0], 1)
        max_array = params['maximum'].reshape(params.shape[0], 1)
        # Check bounds => boolean array where True is out of bounds
        # All NaN values will evaluate to false in a comparison;
        # No special handling needed for unbounded values
        # bounds_mask is 2 dimensional
        bounds_mask = (data < min_array) + (data > max_array)
        counter = 0
        while bounds_mask.sum() > 0:
            # This isn't the most efficient, but random number generation
            # is fast, so pass the whole params. The problem is that data is
            # a large multi-dimensional array, and there is no easy way to pick
            # the values that should be regenerated. Testing shows that this
            # approach increases execution speed linearly as a function of the
            # most restrictive bounds, as all the other bounds are satisfied
            # during iteration for the worst-case scenario. As this approach is
            # O(n), and the time for random number generation is << time for
            # solving the linear system, we don't try to find a clever way
            # around this inefficiency. See stats/tests/uncertainty.py -
            # UncertaintyTestCase - test_random_timing for a timing test.
            data[bounds_mask] = cls.random_variables(params,
                                                     size, seeded_random)[bounds_mask]
            bounds_mask = (data < min_array) + (data > max_array)

            counter += 1
            if counter >= maximum_iterations:
                raise MaximumIterationsError
        return data

    @classmethod
    def random_variables(cls, params, size, seeded_random=None):
        """Generate random variables for the given uncertainty. Should **not check** to ensure that random samples are with the (minimum, maximum bounds). Bounds checking is provided by the `bounded_random_variables` class method.

.. rubric:: Inputs

* params : A :ref:`params-array`.
* size : Integer. The number of values to draw from each distribution in `params`.
* seeded_random : Integer. Optional.

.. rubric:: Output

An array of random values, with dimensions `params` rows by `size`."""
        raise NotImplementedError

    # Used for Latin Hypercube Monte Carlo ###
    @classmethod
    def ppf(cls, params, percentages):
        """Return percent point function (inverse of CDF, e.g. value in distribution where x percent of the distribution is less than value) for various distributions.

.. rubric:: Inputs

* params : A :ref:`params-array`.
* percentages : An array of percentages, bounded on (0,1). Each row in `percentages` corresponds to a row in `params`.

.. rubric:: Output

An array of values within the ranges of each distribtion, with `params` rows and `percentages` columns."""
        percentages = cls.check_ppf_inputs(params, percentages)
        raise NotImplementedError

    @classmethod
    def cdf(cls, params, vector):
        """Used when a distribution is bounded, to determine where to begin or end the percentages used in calculating hypercube sampling space.

.. rubric:: Inputs

* params : A :ref:`params-array`.
* vector : A array of values taken from the uncertainty distributions, with **one row** or the **same number** of rows as `params`.

.. rubric:: Output

An array of cumulative densities, bounded on (0,1), with `params` rows and `vector` columns."""
        vector = cls.check_cdf_inputs(params, vector)
        raise NotImplementedError

    # Used for graphing ###
    @classmethod
    @one_row_params_array
    def statistics(cls, params):
        """Build a dictionary of mean, mode, median, and 95% confidence interval upper and lower values.

.. rubric:: Inputs

* params : A one-row :ref:`params-array`.

.. rubric:: Output

{'mean': mean value, 'mode': mode value, 'median': median value, 'upper': upper limit value, 'lower': lower limit value}. All values should be floats (not single-element arrays). Parameters that are not defined should be returned `None`, not omitted.
        """
        return {'mean': params['loc'], 'mode': None, 'median': None,
                'upper': None, 'lower': None}

    @classmethod
    @one_row_params_array
    def pdf(cls, params, xs=None):
        """Provide a standard interface to calculate the probability distribution function of a uncertainty distribution. Default is `cls.default_number_points_in_pdf` points between min to max range if bounds are present, or `cls.standard_deviations_in_default_range` standard distributions.

.. rubric:: Inputs

* params : A one-row :ref:`params-array`.
* xs : Optional. A one-dimensional numpy array of input values.

.. rubric:: Output

.. important:: The output format for PDF is different than CDF or PPF.

A tuple of a vactor x values and a vector of y values. Y values are a one-dimensional array of probability densities, bounded on (0,1), with length `xs`, if provided, or `cls.default_number_points_in_pdf`."""
        raise NotImplementedError


class BoundedUncertaintyBase(UncertaintyBase):

    """An uncertainty distribution where minimum and maximum bounds are required. No bounds checking is required for these distributions, as bounds are integral inputs into the sample space generator."""
    @classmethod
    def validate(cls, params):
        super(BoundedUncertaintyBase, cls).validate(params)
        if np.isnan(params['minimum']).sum() or np.isnan(params['maximum']).sum():
            raise ImproperBoundsError("This distribution require minimum and maximum values.")

    @classmethod
    def rescale(cls, params):
        """Rescale params to a (0,1) interval. Return adjusted_means and scale. Needed because SciPy assumes a (0,1) interval for many distributions."""
        scale = (params['maximum'] - params['minimum'])
        adjusted_means = (params['loc'] - params['minimum']) / scale
        return adjusted_means, scale

    @classmethod
    def bounded_random_variables(cls, params, size, seeded_random=None,
                                 maximum_iterations=None):
        """No bounds checking because the bounds do not exclude any of the distribution."""
        return cls.random_variables(params, size, seeded_random)

    @classmethod
    @one_row_params_array
    def check_bounds_reasonableness(cls, params):
        """Always true because the bounds do not exclude any of the distribution."""
        return



from __future__ import division
from ...distributions import StudentsTUncertainty
from ...errors import InvalidParamsError
from ..base import UncertaintyTestCase
import numpy as np


class StudentsTTestCase(UncertaintyTestCase):

    def pretty_close(self, a, b):
        if b == 0:
            self.assertTrue(a - 0.1 < b < a + 0.1)
        else:
            self.assertTrue(0.95 * a < b < 1.05 * a)

    def test_loc_and_scale_nan(self):
        params = self.make_params_array()
        params['shape'] = 1
        sample = StudentsTUncertainty.random_variables(params, 5000)
        self.pretty_close(np.median(sample), 0)

    def test_loc_matters(self):
        params = self.make_params_array()
        params['shape'] = 1
        params['loc'] = 10
        sample = StudentsTUncertainty.random_variables(params, 1000)
        self.pretty_close(np.median(sample), 10)

    def test_scale_matters(self):
        params = self.make_params_array()
        params['shape'] = 1
        sample_1 = StudentsTUncertainty.random_variables(params, 5000)
        params['scale'] = 5
        sample_2 = StudentsTUncertainty.random_variables(params, 5000)
        self.assertTrue(np.std(sample_1) < np.std(sample_2))

    def test_random_variables(self):
        params = self.make_params_array()
        params['shape'] = 5
        sample = StudentsTUncertainty.random_variables(params, 20000)
        # nu / (nu - 2) if nu > 2
        expected_variance = 5. / 3
        self.pretty_close(np.var(sample), expected_variance)

    def test_scale_validation(self):
        params = self.make_params_array()
        params['shape'] = 1
        # NaN is OK
        StudentsTUncertainty.validate(params)
        # > 0 is OK
        params['scale'] = 1
        StudentsTUncertainty.validate(params)
        # <= 0 is not
        params['scale'] = 0
        self.assertRaises(
            InvalidParamsError,
            StudentsTUncertainty.validate,
            params
        )
        params['scale'] = -1
        self.assertRaises(
            InvalidParamsError,
            StudentsTUncertainty.validate,
            params
        )

    def test_shape_validation(self):
        params = self.make_params_array()
        params['shape'] = 1
        # > 0 is OK
        params['shape'] = 1
        StudentsTUncertainty.validate(params)
        # <= 0 is not
        params['shape'] = 0
        self.assertRaises(
            InvalidParamsError,
            StudentsTUncertainty.validate,
            params
        )
        params['shape'] = -1
        self.assertRaises(
            InvalidParamsError,
            StudentsTUncertainty.validate,
            params
        )



from .distributions import *
import warnings

DISTRIBUTIONS = (
    BernoulliUncertainty,
    BetaUncertainty,
    DiscreteUniform,
    GammaUncertainty,
    GeneralizedExtremeValueUncertainty,
    LognormalUncertainty,
    NormalUncertainty,
    NoUncertainty,
    StudentsTUncertainty,
    TriangularUncertainty,
    UndefinedUncertainty,
    UniformUncertainty,
    WeibullUncertainty,
)


class UncertaintyChoices(object):

    """An container for uncertainty distributions."""

    def __init__(self):
        # Sorted by id
        self.choices = sorted(DISTRIBUTIONS, key=lambda x: x.id)
        self.check_id_uniqueness()

    def check_id_uniqueness(self):
        self.id_dict = {}
        for dist in self.choices:
            if dist.id in self.id_dict:
                raise ValueError("Uncertainty id %i is already in use by %s" %
                                (dist.id, self.id_dict[dist.id]))
            self.id_dict[dist.id] = dist

    def __iter__(self):
        return iter(self.choices)

    def __getitem__(self, index):
        return self.id_dict[index]

    def __len__(self):
        return len(self.id_dict)

    def __contains__(self, choice):
        return choice in self.choices

    def add(self, distribution):
        if not hasattr(distribution, "id") and isinstance(distributions.id, int):
            raise ValueError("Uncertainty distributions must have integer `id` attribute.")
        if distribution.id in self.id_dict:
            warnings.warn("ERROR: This distribution (id %s) is already present!" % distribution.id)
            return
        self.choices.append(distribution)
        self.id_dict[distribution.id] = distribution

uncertainty_choices = UncertaintyChoices()



from __future__ import division
from ..errors import InvalidParamsError
from ..utils import one_row_params_array
from .base import UncertaintyBase
from scipy import stats
import numpy as np


class NormalUncertainty(UncertaintyBase):
    id = 3
    description = "Normal uncertainty"

    @classmethod
    def validate(cls, params):
        if np.isnan(params['scale']).sum() or (params['scale'] <= 0).sum():
            raise InvalidParamsError(
                "Real, positive scale (sigma) values are required"
                " for normal uncertainties."
            )
        if np.isnan(params['loc']).sum():
            raise InvalidParamsError(
                "Real loc (mu) values are required for normal uncertainties."
            )

    @classmethod
    def random_variables(cls, params, size, seeded_random=None):
        if not seeded_random:
            seeded_random = np.random
        return seeded_random.normal(
            params['loc'],
            params['scale'],
            size=(size, params.shape[0])).T

    @classmethod
    def cdf(cls, params, vector):
        vector = cls.check_2d_inputs(params, vector)
        results = np.zeros(vector.shape)
        for row in range(params.shape[0]):
            results[row, :] = stats.norm.cdf(
                vector[row, :],
                loc=params['loc'][row],
                scale=params['scale'][row]
            )
        return results

    @classmethod
    def ppf(cls, params, percentages):
        percentages = cls.check_2d_inputs(params, percentages)
        results = np.zeros(percentages.shape)
        for row in range(percentages.shape[0]):
            results[row, :] = stats.norm.ppf(
                percentages[row, :],
                loc=params['loc'][row],
                scale=params['scale'][row]
            )
        return results

    @classmethod
    @one_row_params_array
    def statistics(cls, params):
        return {
            'mean': float(params['loc']),
            'mode': float(params['loc']),
            'median': float(params['loc']),
            'lower': float(params['loc'] - 2 * params['scale']),
            'upper': float(params['loc'] + 2 * params['scale'])
        }

    @classmethod
    @one_row_params_array
    def pdf(cls, params, xs=None):
        if xs is None:
            if np.isnan(params['minimum']):
                lower = params['loc'] - params['scale'] * \
                    cls.standard_deviations_in_default_range
            else:
                lower = params['minimum']
            if np.isnan(params['maximum']):
                upper = params['loc'] + params['scale'] * \
                    cls.standard_deviations_in_default_range
            else:
                upper = params['maximum']
            xs = np.arange(
                lower,
                upper,
                (upper - lower) / cls.default_number_points_in_pdf
            )

        ys = stats.norm.pdf(xs, params['loc'], params['scale'])
        return xs, ys.reshape(ys.shape[1])



from __future__ import division
from base import UncertaintyBase
from ..errors import InvalidParamsError
import numpy as np


class GeneralizedExtremeValueUncertainty(UncertaintyBase):

    r"""
The generalized extreme value uncertainty distribution has the cumulative distribution function (the PDF is even more complicated):

.. math:: f(x;\mu,\sigma,\xi) = \exp\left\{-\left[1+\xi\left(\frac{x-\mu}{\sigma}\right)\right]^{-1/\xi}\right\}

In our implementation, :math:`\mu` is ``location``, :math:`\sigma` is ``scale``, and :math:`\xi`  is ``shape``.

See http://en.wikipedia.org/wiki/Generalized_extreme_value_distribution.
    """
    id = 11
    description = "Generalized extreme value uncertainty"

    @classmethod
    def validate(cls, params):
        if np.isnan(params['loc']).sum():
            raise InvalidParamsError(
                "Real ``mu`` values needed for generalized extreme value.")
        if (params['scale'] <= 0).sum():
            raise InvalidParamsError(
                "Real, positive ``sigma`` values need for generalized extreme value.")
        if (params['shape'] != 0).sum():
            raise InvalidParamsError(
                "Non-zero ``xi`` values are not yet supported.")

    @classmethod
    def random_variables(cls, params, size, seeded_random=None, **kwargs):
        if seeded_random is None:
            seeded_random = np.random
        data = seeded_random.gumbel(
            loc=params['loc'],
            scale=params['scale'],
            size=(size, params.shape[0])
        ).T
        return data
